import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.interfaces.transformer import Transformer


class EliminateRedundantShapePattern(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [
            Pattern_1,
            Pattern_2,
            Pattern_4,
            Pattern_5,
            Pattern_6,
            Pattern_7,
            Pattern_3,
            Pattern_8,
        ]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Flatten/Squeeze --> Unsqueeze --> next
    to
        prev --> (   ) --> next
    if prev.output[0].shape == next.input[0].shape
    """

    pattern_to_match = ['Flatten/Squeeze', 'Unsqueeze']

    def pattern_matching(self, base_node):
        inputs = base_node.input

        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return inputs

        if not self.pattern_condition_checker(matched_nodes):
            return inputs

        top_node = matched_nodes[0]

        self.transform_to_eliminate(matched_nodes, top_node.input[0])
        return top_node.input

    def pattern_condition_checker(self, nodes_to_check):
        top_node = nodes_to_check[0]
        base_node = nodes_to_check[-1]

        if self.is_same_shape(top_node.input[0], base_node.output[0]):
            return True
        return False


class Pattern_2(Pattern_1):
    """
    transform
        prev --> Reshape --> Flatten/Squeeze --> Unsqueeze --> next
    to
        prev --> (   ) --> next
    if prev.output[0].shape == next.input[0].shape
    """

    pattern_to_match = ['Reshape', 'Flatten/Squeeze', 'Unsqueeze']


class Pattern_3(Pattern_1):
    """
    transform
        prev --> Reshape --> next
    to
        prev --> (   ) --> next
    if prev.output[0].shape == next.input[0].shape
    """

    pattern_to_match = ['Reshape']


class Pattern_4(Pattern_1):
    """
    transform
        prev --> Reshape --> Expand --> Expand --> Reshape --> next

    to
        prev --> (   ) --> next
    if prev.output[0].shape == next.input[0].shape
    """

    pattern_to_match = ['Reshape', 'Expand', 'Expand', 'Reshape']


class Pattern_5(Pattern_1):
    """
    transform
        prev --> Reshape --> Expand --> Reshape --> next
    to
        prev --> (   ) --> next
    if prev.output[0].shape == next.input[0].shape
    """

    pattern_to_match = ['Reshape', 'Expand', 'Reshape']


class Pattern_6(Pattern_1):
    """
    transform
        prev --> Reshape --> Reshape --> next
    to
        prev --> (   ) --> next
    if prev.output[0].shape == next.input[0].shape
    """

    pattern_to_match = ['Reshape', 'Reshape']


class Pattern_7(Pattern_1):
    """
    transform
        prev --> Reshape --> Reshape --> Reshape --> next
    to
        prev --> (   ) --> next
    if prev.output[0].shape == next.input[0].shape
    """

    pattern_to_match = ['Reshape', 'Reshape', 'Reshape']


class Pattern_8(Pattern_1):
    """
    transform
        prev --> Expand --> next
    to
        prev --> (   ) --> next
    if prev.output[0].shape == next.input[0].shape
    """

    pattern_to_match = ['Expand']



import onnx
from onnx import version_converter

from furiosa.quantizer.frontend.onnx import __OPSET_VERSION__
from furiosa.quantizer.frontend.onnx.transformer.utils import (
    eliminate_initializer_from_graph_input,
    include_initializer_to_graph_input,
)
from furiosa.quantizer.frontend.onnx.utils.check_model import check_model
from furiosa.quantizer.interfaces.transformer import Transformer


class CheckVersion(Transformer[onnx.ModelProto]):
    """Convert an ONNX model to ONNX opset 12"""

    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        # https://github.com/onnx/onnx/issues/2873#issuecomment-652541006
        #
        # > There is an underlying issue in version converter. It relies on the
        # > C++ IR which I believe has not been updated after IR v3. Because of
        # > this I think it expects initializers also be added as graph inputs.
        # > If you try to change the version of your model to IRv3 or create a
        # > model with initializers also as inputs then I think this will work.
        model = include_initializer_to_graph_input(model)

        version = int(model.opset_import[0].version)
        if version != __OPSET_VERSION__:
            try:
                model = version_converter.convert_version(model, __OPSET_VERSION__)
                check_model(model, check_runnable=False)
            except Exception as exc:
                raise NotImplementedError(
                    f"can't convert the model (ONNX opset {version}) to ONNX opset {__OPSET_VERSION__}"
                ) from exc
        model = eliminate_initializer_from_graph_input(model)
        return model



import onnx
from onnx.helper import make_node

from furiosa.quantizer.frontend.onnx.transformer import utils
from furiosa.quantizer.frontend.onnx.utils.check_model import check_model
from furiosa.quantizer.interfaces.transformer import Transformer


class Convert2dSumToAdd(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        optimized_nodes = []
        for node in model.graph.node:
            if node.op_type != 'Sum':
                optimized_nodes.append(node)
                continue

            if len(node.input) != 2:
                optimized_nodes.append(node)
                continue

            new_node = make_node(
                'Add', inputs=[node.input[0], node.input[1]], outputs=[node.output[0]]
            )

            optimized_nodes.append(new_node)

        # remove duplicate node(s) in optimized nodes
        seen = []
        for op_node in optimized_nodes:
            if op_node in seen:
                continue
            seen.append(op_node)
        optimized_nodes = seen

        model = utils.rebuild_model(model, optimized_nodes)
        check_model(model)

        return model



from collections import defaultdict
import itertools
from typing import Callable, Dict, Iterable, List, Optional, Tuple

import numpy as np
import onnx

__DOMAIN__ = ''
__OPSET_VERSION__ = 12

from furiosa.quantizer.frontend.onnx import calibrate
from furiosa.quantizer.frontend.onnx.quantizer import quantizer
from furiosa.quantizer.frontend.onnx.transformer.convert_conv1d_to_conv2d import (
    ConvertConv1dToConv2d,
)
from furiosa.quantizer.frontend.onnx.transformer.eliminate_redundant_shape_pattern import (
    EliminateRedundantShapePattern,
)
from furiosa.quantizer.frontend.onnx.transformer.fuse_batchnorm import FuseBatchNorm
from furiosa.quantizer.frontend.onnx.transformer.fuse_conv import FuseConv
from furiosa.quantizer.frontend.onnx.transformer.fuse_depth_to_space import FuseDepthToSpace
from furiosa.quantizer.frontend.onnx.transformer.fuse_gather_matmul import FuseGatherMatMul
from furiosa.quantizer.frontend.onnx.transformer.fuse_gelu import FuseGELU
from furiosa.quantizer.frontend.onnx.transformer.fuse_layer_normalization import (
    FuseLayerNormalization,
)
from furiosa.quantizer.frontend.onnx.transformer.fuse_lp_normalization import FuseLpNormalization
from furiosa.quantizer.frontend.onnx.transformer.fuse_pad import FusePad
from furiosa.quantizer.frontend.onnx.transformer.fuse_redundant_reshape_pattern import (
    FuseRedundantReshapePattern,
)
from furiosa.quantizer.frontend.onnx.transformer.polish_model import PolishModel
from furiosa.quantizer.frontend.onnx.utils.version_checker import CheckVersion

_CONV = "Conv"
_DEQUANTIZE_LINEAR = "DequantizeLinear"
_MAT_MUL = "MatMul"
_QUANTIZE_LINEAR = "QuantizeLinear"
_Q_LINEAR_CONV = "QLinearConv"
_Q_LINEAR_MAT_MUL = "QLinearMatMul"


def optimize_model(
    model: onnx.ModelProto, input_shapes: Optional[Dict[str, List[int]]] = None
) -> onnx.ModelProto:
    model = _transform([CheckVersion().transform], model)
    model = _transform([PolishModel(input_shapes).transform], model)

    # TODO check if graph_transform should apply.
    model = _transform([_reify], model)
    return model


def quantize(
    model: onnx.ModelProto,
    per_channel: bool,
    static: bool,
    mode: quantizer.QuantizationMode,
    dynamic_ranges: Dict[str, Tuple[float, float]],
) -> onnx.ModelProto:
    return quantizer.FuriosaONNXQuantizer(
        model, per_channel, static, mode, dynamic_ranges
    ).quantize()


def post_training_quantize(
    model: onnx.ModelProto,
    dataset: Iterable[Dict[str, np.ndarray]],
    per_channel: bool = True,
    check_idempotency: bool = False,
) -> onnx.ModelProto:
    """Post-training-quantizes an ONNX model with a calibration dataset.

    Args:
        model: An ONNX model to quantize.
        dataset: A calibration dataset.
        per_channel: If per_channel is True, Conv's filters are
          per-channel quantized. Otherwise, they are per-tensor
          quantized.
    Returns:
        An ONNX model post-training-quantized with the calibration
        dataset.
    """
    if check_idempotency:
        if _is_fully_quantized(model):
            return model
        if any(node.op_type in [_DEQUANTIZE_LINEAR, _QUANTIZE_LINEAR] for node in model.graph.node):
            raise ValueError(
                "an ONNX model with DequantizeLinear or QuantizeLinear is not supported yet."
            )

    model = optimize_model(model)
    ranges = calibrate.calibrate(model, dataset)
    return quantize(model, per_channel, True, quantizer.QuantizationMode.DFG, ranges)


def post_training_quantization_with_random_calibration(
    model: onnx.ModelProto,
    per_channel: bool,
    static: bool,
    mode: quantizer.QuantizationMode,
    num_data: int = 8,
    check_idempotency: bool = False,
) -> onnx.ModelProto:
    if not static:
        raise Exception("Currently only supports static quantization.")

    if check_idempotency:
        if _is_fully_quantized(model):
            return model
        if any(node.op_type in [_DEQUANTIZE_LINEAR, _QUANTIZE_LINEAR] for node in model.graph.node):
            raise ValueError(
                "an ONNX model with DequantizeLinear or QuantizeLinear is not supported yet."
            )

    model = optimize_model(model)
    dynamic_ranges = calibrate.calibrate_with_random_data(model, num_data)
    return quantize(model, per_channel, static, mode, dynamic_ranges)


def parse_onnx_graph(
    model: onnx.ModelProto,
) -> Tuple[
    Dict[str, onnx.ValueInfoProto], Dict[str, onnx.NodeProto], Dict[str, List[onnx.NodeProto]]
]:
    model = onnx.shape_inference.infer_shapes(model)

    value_infos = {
        value_info.name: value_info
        for value_info in itertools.chain(
            model.graph.input, model.graph.output, model.graph.value_info
        )
    }
    # append graph.initializer's value_info to value_infos
    value_infos.update(
        (init.name, onnx.helper.make_tensor_value_info(init.name, init.data_type, init.dims))
        for init in model.graph.initializer
    )
    assert all(
        tensor in value_infos
        for node in model.graph.node
        for tensor in itertools.chain(node.input, node.output)
    ), "ONNX Shape inference did not work well on some tensor(s). All value_infos must be inferenced to proceed."
    producer = {tensor: node for node in model.graph.node for tensor in node.output}
    consumers = defaultdict(list)
    for node in model.graph.node:
        for tensor in node.input:
            consumers[tensor].append(node)

    return value_infos, producer, consumers


def _transform(
    transformers: List[Callable[[onnx.ModelProto], onnx.ModelProto]], model: onnx.ModelProto
) -> onnx.ModelProto:
    for transform in transformers:
        model = transform(model)
    return model


def _reify(model: onnx.ModelProto) -> onnx.ModelProto:
    transformers = [
        ConvertConv1dToConv2d().transform,
        FuseConv().transform,
        FusePad().transform,
        FuseBatchNorm().transform,
        FuseDepthToSpace().transform,
        FuseGELU().transform,
        FuseLayerNormalization().transform,
        FuseLpNormalization().transform,
        FuseRedundantReshapePattern().transform,
        FuseGatherMatMul().transform,
        EliminateRedundantShapePattern().transform,
    ]
    return _transform(transformers, model)


def _is_fully_quantized(model: onnx.ModelProto) -> bool:
    """Return `True` if the ONNX `graph` is already quantized."""
    return _is_fully_quantized_in_dfg_mode(
        model.graph, *parse_onnx_graph(model)
    ) or _is_fully_quantized_in_fake_quant_mode(model.graph, *parse_onnx_graph(model))


def _is_fully_quantized_in_dfg_mode(
    graph: onnx.GraphProto,
    value_infos: Dict[str, onnx.ValueInfoProto],
    producer: Dict[str, onnx.NodeProto],
    consumers: Dict[str, List[onnx.NodeProto]],
) -> bool:
    return all(
        node.op_type in [_DEQUANTIZE_LINEAR, _QUANTIZE_LINEAR, _Q_LINEAR_CONV, _Q_LINEAR_MAT_MUL]
        or (
            node.op_type not in [_CONV, _MAT_MUL]
            and _is_sandwiched(node, value_infos, producer, consumers)
        )
        for node in graph.node
    )


def _is_fully_quantized_in_fake_quant_mode(
    graph: onnx.GraphProto,
    value_infos: Dict[str, onnx.ValueInfoProto],
    producer: Dict[str, onnx.NodeProto],
    consumers: Dict[str, List[onnx.NodeProto]],
) -> bool:
    return all(
        node.op_type in [_DEQUANTIZE_LINEAR, _QUANTIZE_LINEAR]
        or _is_sandwiched(node, value_infos, producer, consumers)
        for node in graph.node
    )


def _is_sandwiched(
    node: onnx.NodeProto,
    value_infos: Dict[str, onnx.ValueInfoProto],
    producer: Dict[str, onnx.NodeProto],
    consumers: Dict[str, List[onnx.NodeProto]],
) -> bool:
    return all(
        value_infos[tensor].type.tensor_type.elem_type != onnx.TensorProto.FLOAT
        or (tensor in producer and producer[tensor].op_type == _DEQUANTIZE_LINEAR)
        for tensor in node.input
    ) and all(
        value_infos[tensor].type.tensor_type.elem_type != onnx.TensorProto.FLOAT
        or (
            tensor in consumers
            and all(consumer.op_type == _QUANTIZE_LINEAR for consumer in consumers[tensor])
        )
        for tensor in node.output
    )



from typing import Optional

import numpy as np
import onnx

from furiosa.quantizer.frontend.onnx.transformer import utils
from furiosa.quantizer.frontend.onnx.utils.check_model import check_model


class DFGImportable:
    def __init__(self, model, raw_data):
        copy_model = onnx.ModelProto()
        copy_model.CopyFrom(model)
        self.model = copy_model
        self.raw_data = raw_data

        self.node = {node.name: node for node in self.model.graph.node}
        self.node_by_output = {
            node_output: node for node in self.model.graph.node for node_output in node.output
        }
        self.node_by_input = {
            node_input: node for node in self.model.graph.node for node_input in node.input
        }
        self.initializer = {init.name: init for init in self.model.graph.initializer}
        self.graph_input = {vi.name: vi for vi in self.model.graph.input}
        self.value_info = {vi.name: vi for vi in self.model.graph.value_info}

    def transform(self):
        self.remove_quantizelinear_operator_with_initializer()
        self.transform_to_integer_arithmetic_operator()

        return self.model

    def remove_quantizelinear_operator_with_initializer(self):
        new_nodes = []
        rm_nodes = []
        for node in self.model.graph.node:
            if node.op_type != 'QuantizeLinear':
                new_nodes.append(node)
                continue
            if node.input[0] not in self.initializer:
                new_nodes.append(node)
                continue

            # node.input[0] to be removed from model.graph.initializer
            init = self.initializer.pop(node.input[0])

            # quantize initializer
            s = self.initializer[node.input[1]]
            zp = self.initializer[node.input[2]]

            dequant_linear_output = node.output[0].split("_quantized")[0] + "_dequantized"
            middle_node = self.node_by_input[dequant_linear_output]

            # gives output_channel_axis for per-channel quantized Conv/ConvTranspose weight/bias
            rank = len(init.dims)
            if rank > 1:
                if middle_node.op_type == "Conv":
                    output_channel_axis = 0
                elif middle_node.op_type == 'ConvTranspose':
                    output_channel_axis = 1
                else:
                    output_channel_axis = None
            else:
                output_channel_axis = None

            quantized_data = self._quantize_data(init, s, zp, output_channel_axis)

            # node.output[0] to be updated to model.graph.initializer instead
            flattened = quantized_data.flatten()
            if self.raw_data:
                flattened = flattened.tobytes()

            self.initializer.update(
                {
                    node.output[0]: onnx.helper.make_tensor(
                        node.output[0], zp.data_type, init.dims, flattened, raw=self.raw_data
                    )
                }
            )

            # node.output[0] to be removed from model.graph.value_info
            self.value_info.pop(node.output[0])

            rm_nodes.append(node)

        self.model = utils.rebuild_model(
            model=self.model,
            new_nodes=[node for node in new_nodes if node not in rm_nodes],
            eliminate=False,
            renaming=False,
        )

        self._update_graph_field(field='initializer', proto=self.initializer.values())
        self._update_graph_field(field='value_info', proto=self.value_info.values())
        self._update_graph_field(field='input', proto=self.graph_input.values())

        check_model(self.model, check_runnable=False)

    def transform_to_integer_arithmetic_operator(self):
        new_nodes = []
        rm_nodes = []

        for node in self.model.graph.node:
            if node.op_type not in ['Conv', 'MatMul']:
                new_nodes.append(node)
                continue

            node_i0 = self.node_by_output[node.input[0]]
            node_i1 = self.node_by_output[node.input[1]]
            node_o0 = self.node_by_input[node.output[0]]
            rm_nodes.extend([node_i0, node_i1, node_o0])

            # When a tensor is used as model's output, it is not in the value_info.
            # 'None' argument to handle the case.
            self.value_info.pop(node_i0.output[0], None)
            self.value_info.pop(node_i1.output[0], None)
            self.value_info.pop(node_o0.input[0], None)

            node_i2 = None
            if len(node.input) == 3:
                node_i2 = self.node_by_output[node.input[2]]
                rm_nodes.append(node_i2)

                self.value_info.pop(node_i2.output[0], None)

            rm_nodes.extend([node])
            new_nodes.append(
                self._make_integer_arithmetic_operator(node, node_i0, node_i1, node_o0, node_i2)
            )

        self.model = utils.rebuild_model(
            model=self.model,
            new_nodes=[node for node in new_nodes if node not in rm_nodes],
            eliminate=False,
        )

        self._update_graph_field(field='value_info', proto=self.value_info.values())

        check_model(self.model, check_runnable=False)

    @staticmethod
    def _make_integer_arithmetic_operator(node, node_i0, node_i1, node_o0, node_i2=None):
        quant_op_type = f'QLinear{node.op_type}'

        quant_inputs = [*node_i0.input, *node_i1.input, *node_o0.input[1:]]

        if node_i2:
            quant_inputs.append(node_i2.input[0])

        quant_outputs = [node_o0.output[0]]

        attr_kwargs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in node.attribute}

        quant_node = onnx.helper.make_node(
            quant_op_type, quant_inputs, quant_outputs, **attr_kwargs
        )

        return quant_node

    @staticmethod
    def _quantize_data(
        data: onnx.TensorProto,
        scale: onnx.TensorProto,
        zero_point: onnx.TensorProto,
        axis: Optional[int] = None,
    ) -> np.ndarray:
        data_arr = np.atleast_1d(onnx.numpy_helper.to_array(data)).astype(np.float32)
        scale_arr = np.atleast_1d(onnx.numpy_helper.to_array(scale)).astype(np.float32)
        zero_point_arr = np.atleast_1d(onnx.numpy_helper.to_array(zero_point)).astype(np.float32)

        if axis is not None:
            new_axes = list(range(data_arr.ndim))
            new_axes.pop(axis)
            scale_arr = np.expand_dims(scale_arr, axis=new_axes)
            zero_point_arr = np.expand_dims(zero_point_arr, axis=new_axes)

        quantized_data = np.round(data_arr / scale_arr) + zero_point_arr

        np_dtype_info = np.iinfo(onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[zero_point.data_type])
        np_dtype_info_min = np_dtype_info.min
        np_dtype_info_max = np_dtype_info.max

        return np.clip(quantized_data, np_dtype_info_min, np_dtype_info_max).astype(
            onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[zero_point.data_type]
        )

    def _update_graph_field(self, field, proto):
        self.model.graph.ClearField(field)
        getattr(self.model.graph, field).extend(proto)


class ONNXRuntimeExecutable(DFGImportable):
    def transform(self):

        self._remove_quant_dequantlinear_operator_with_initializer()

        return self.model

    def _remove_quant_dequantlinear_operator_with_initializer(self):
        rm_nodes = []
        new_nodes = []
        for node in self.model.graph.node:
            if node.op_type != 'QuantizeLinear':
                new_nodes.append(node)
                continue

            if node.input[0] not in self.initializer:
                new_nodes.append(node)
                continue

            rm_nodes.append(node)
            rm_nodes.extend(
                [
                    dequant_node
                    for dequant_node in self.model.graph.node
                    if dequant_node.op_type == 'DequantizeLinear'
                    if dequant_node.input[0] == node.output[0]
                ]
            )

        for node in self.model.graph.node:
            if node.op_type in ('QuantizeLinear', 'DequantizeLinear'):
                continue

            for idx, node_input in enumerate(node.input):
                if '_dequantized' not in node_input:
                    continue

                init_name = node_input.split('_dequantized')[0]
                if init_name not in self.initializer:
                    continue

                node.input[idx] = init_name + '_fake_quantized'

                init = self.initializer.pop(init_name)
                s = self.initializer.pop(init_name + '_scale')
                zp = self.initializer.pop(init_name + '_zero_point')

                # gives output_channel_axis for per-channel quantized Conv/ConvTranspose weight/bias
                rank = len(init.dims)
                if rank > 1:
                    if node.op_type == "Conv":
                        output_channel_axis = 0
                    elif node.op_type == 'ConvTranspose':
                        output_channel_axis = 1
                    else:
                        output_channel_axis = None
                else:
                    output_channel_axis = None

                fake_quantized_data = self._fake_quantize_data(init, s, zp, output_channel_axis)

                flattened = fake_quantized_data.flatten()
                if self.raw_data:
                    flattened = flattened.tobytes()
                self.initializer.update(
                    {
                        init_name
                        + '_fake_quantized': onnx.helper.make_tensor(
                            name=init_name + '_fake_quantized',
                            data_type=onnx.TensorProto.FLOAT,
                            dims=init.dims,
                            vals=flattened,
                            raw=self.raw_data,
                        )
                    }
                )

        self.model = utils.rebuild_model(
            model=self.model,
            new_nodes=[node for node in new_nodes if node not in rm_nodes],
            eliminate=True,
        )
        self._update_graph_field(field='initializer', proto=self.initializer.values())

        check_model(self.model, check_runnable=True)

    @staticmethod
    def _dequantize_data(
        data: onnx.TensorProto,
        scale: onnx.TensorProto,
        zero_point: onnx.TensorProto,
        axis: Optional[int] = None,
    ) -> np.ndarray:
        data_arr = np.atleast_1d(onnx.numpy_helper.to_array(data)).astype(np.float32)
        scale_arr = np.atleast_1d(onnx.numpy_helper.to_array(scale)).astype(np.float32)
        zero_point_arr = np.atleast_1d(onnx.numpy_helper.to_array(zero_point)).astype(np.float32)

        if axis is not None:
            new_axes = list(range(data_arr.ndim))
            new_axes.pop(axis)
            scale_arr = np.expand_dims(scale_arr, axis=new_axes)
            zero_point_arr = np.expand_dims(zero_point_arr, axis=new_axes)

        return (data_arr - zero_point_arr) * scale_arr

    def _fake_quantize_data(
        self,
        data: onnx.TensorProto,
        scale: onnx.TensorProto,
        zero_point: onnx.TensorProto,
        axis: Optional[int] = None,
    ) -> np.ndarray:
        quantized_data = self._quantize_data(data, scale, zero_point, axis)
        flattened = quantized_data.flatten()
        if self.raw_data:
            flattened = flattened.tobytes()
        dequantized_data = self._dequantize_data(
            onnx.helper.make_tensor(
                name=data.name,
                data_type=zero_point.data_type,
                dims=data.dims,
                vals=flattened,
                raw=self.raw_data,
            ),
            scale,
            zero_point,
            axis,
        )

        return dequantized_data



from abc import ABC, abstractmethod
from typing import List

import numpy as np
import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.interfaces.transformer import Transformer


class EliminateClipper(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [
            Pattern_1,
            Pattern_2,
            Pattern_3,
            Pattern_4,
            Pattern_5,
            Pattern_6,
        ]:
            transformer = transformer(model)
            model = transformer.transform()
        return model


class ClipperElimination(ONNXTransformer, ABC):
    """
    This class contains methods commonly used in various EliminateClipper patterns

    Eliminate clippers if
        1. a pattern given should be matched.
        2. every consumer's output quantization parameters of the previous node of clipper should be mutually identical.
    """

    # EliminateClipper().transform() is called on QDQ graph.
    # QDQ graph will apply QuantizeLinear - DequantizeLinear to bias of Conv operators.
    # bias of QLinearConv operator should be i32, so QDQ graph will have i32 quantization information for bias.
    # However, QuantizeLinear operator of onnx are only defined on i8 quantization parmeter, so temporarily, QDQ graph is not valid.
    check_runnable = False

    @property
    @abstractmethod
    def pattern_to_match(self) -> List[str]:
        pass

    def pattern_matching(self, base_node):
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        # cut out DQ(dqlinear)-clip-Q1(qlinear1) in ...-op-Q-DQ-clip-Q1-(DQ1)
        # connect op-Q with DQ1
        op, qlinear, dqlinear, clip, qlinear1 = matched_nodes[-5:]

        self.transform_to_eliminate([dqlinear, clip, qlinear1], qlinear.output[0])
        self.replace_quant_params(qlinear, qlinear1)
        self.remove_clip_qdq(clip)

        return op.input

    def pattern_condition_checker(self, matched_nodes):
        *_, _, qlinear, _, clip, _ = matched_nodes
        return self.check_condition_1(clip) and self.check_condition_2(qlinear)

    def check_condition_1(self, clip):
        assert clip.op_type in ("Clip", "Relu"), repr(clip)
        if clip.op_type == "Relu":
            return True

        if len(clip.input) == 1:
            return True
        if len(clip.input) == 2:
            min_tensor = self.find_prev_node(self.find_prev_node(clip.input[1]).input[0]).input[0]
            return min_tensor in self.initializer_map
        if len(clip.input) == 3:
            min_tensor = self.find_prev_node(self.find_prev_node(clip.input[1]).input[0]).input[0]
            max_tensor = self.find_prev_node(self.find_prev_node(clip.input[2]).input[0]).input[0]
            return min_tensor in self.initializer_map and max_tensor in self.initializer_map
        raise ValueError(clip)

    def check_condition_2(self, qlinear):
        """
        We assume multiple dqlinears(DQs) might follow from qlinear(Q) like in ...-op-Q-DQ-clip-Q1-...
                                                                                      +-DQ_1-clip-Q1_1-...
                                                                                      +-      ...
        Logic below checks the equality of qlinear1s(Q1s)' quantization parameters.
        """
        dqlinears = self.consumer_map[qlinear.output[0]]
        scales = set()
        zero_points = set()
        for dqlinear in dqlinears:
            clip = self.consumer_map[dqlinear.output[0]][0]
            qlinear1 = self.consumer_map[clip.output[0]][0]
            assert qlinear1.op_type == "QuantizeLinear", repr(qlinear1)
            scales.add(float(self.get_initializer_array(qlinear1.input[1])))
            zero_points.add(int(self.get_initializer_array(qlinear1.input[2])))
        return len(scales) == len(zero_points) == 1

    def remove_clip_qdq(self, clip):
        assert clip.op_type in ("Clip", "Relu"), repr(clip)
        for tensor_name in clip.input:
            dqlinear = self.traverse_prev_node(tensor_name, ["DequantizeLinear"])
            qlinear = self.traverse_prev_node(dqlinear.input[0], ["QuantizeLinear"])
            # remove dq-q nodes only if qlinear.input[0] is contant
            if qlinear.input[0] in self.initializer_map:
                self.pop_single_optimizer_map(dqlinear)
                self.pop_single_optimizer_map(qlinear)

    def replace_quant_params(self, qlinear, qlinear1):
        assert qlinear.op_type == "QuantizeLinear", repr(qlinear)
        assert qlinear1.op_type == "QuantizeLinear", repr(qlinear1)

        # replace qlinear's quant params with qlinear1's
        for idx in range(1, 3):
            old_tensor = qlinear.input[idx]
            new_tensor = qlinear1.input[idx]
            if np.array_equal(
                self.get_initializer_array(old_tensor), self.get_initializer_array(new_tensor)
            ):
                # if replacement already happened, then skip.
                continue
            self.initializer_map.pop(old_tensor)
            qlinear.input[idx] = new_tensor


class Pattern_1(ClipperElimination):
    """
    transform
        prev --> Conv --> QuantizeLinear --> DequantizeLinear --> Relu --> QuantizeLinear --> next
    to
        prev --> Conv --> QuantizeLinear --> next
    """

    pattern_to_match = ['Conv', 'QuantizeLinear', 'DequantizeLinear', 'Relu', 'QuantizeLinear']


class Pattern_2(ClipperElimination):
    """
    transform
        prev --> Conv --> QuantizeLinear --> DequantizeLinear --> Clip --> QuantizeLinear --> next
    to
        prev --> Conv --> QuantizeLinear --> next

    if Clip.input[1], Clip.input[2]'s input before fake quantization have initializer
    """

    pattern_to_match = ['Conv', 'QuantizeLinear', 'DequantizeLinear', 'Clip', 'QuantizeLinear']


class Pattern_3(ClipperElimination):
    """
    transform
        prev --> Add --> QuantizeLinear --> DequantizeLinear --> Relu --> QuantizeLinear --> next
    to
        prev --> Add --> QuantizeLinear --> next
    """

    pattern_to_match = ['Add', 'QuantizeLinear', 'DequantizeLinear', 'Relu', 'QuantizeLinear']


class Pattern_4(ClipperElimination):
    """
    transform
        prev --> Add --> QuantizeLinear --> DequantizeLinear --> Clip --> QuantizeLinear --> next
    to
        prev --> Add --> QuantizeLinear --> next

    if Clip.input[1], Clip.input[2]'s input before fake quantization have initializer
    """

    pattern_to_match = ['Add', 'QuantizeLinear', 'DequantizeLinear', 'Clip', 'QuantizeLinear']


class Pattern_5(ClipperElimination):
    """
    transform
        prev --> Conv --> QuantizeLinear --> DequantizeLinear --> Squeeze --> QuantizeLinear --> DequantizeLinear --> Relu --> QuantizeLinear --> next
    to
        prev --> Conv --> QuantizeLinear --> DequantizeLinear --> Squeeze --> QuantizeLinear --> next
    """

    pattern_to_match = [
        'Conv',
        'QuantizeLinear',
        'DequantizeLinear',
        'Squeeze',
        'QuantizeLinear',
        'DequantizeLinear',
        'Relu',
        'QuantizeLinear',
    ]


class Pattern_6(ClipperElimination):
    """
    transform
        prev --> Conv --> QuantizeLinear --> DequantizeLinear --> Squeeze --> QuantizeLinear --> DequantizeLinear --> Clip --> QuantizeLinear --> next
    to
        prev --> Conv --> QuantizeLinear --> DequantizeLinear --> Squeeze --> QuantizeLinear --> next

    if Clip.input[1], Clip.input[2]'s input before fake quantization have initializer
    """

    pattern_to_match = [
        'Conv',
        'QuantizeLinear',
        'DequantizeLinear',
        'Squeeze',
        'QuantizeLinear',
        'DequantizeLinear',
        'Clip',
        'QuantizeLinear',
    ]



from typing import Dict, Iterable

import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.interfaces.transformer import Transformer


class FuseDepthToSpace(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [
            Pattern_1,
        ]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Reshape --> Transpose --> Reshape --> next
    to
        prev --> DepthToSpace --> next

    if Transpose.perm == [0, 1, 4, 2, 5, 3] or == [0, 3, 4, 1, 5, 2]
    """

    def pattern_matching(self, base_node: onnx.NodeProto) -> Iterable[str]:
        pattern_to_match = ['Reshape', 'Transpose', 'Reshape']
        matched_nodes = self.pattern_matcher(base_node, pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not _pattern_condition_checker(matched_nodes):
            return base_node.input

        reshape, *_ = matched_nodes
        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=[
                onnx.helper.make_node(
                    'DepthToSpace',
                    [reshape.input[0]],
                    [base_node.output[0]],
                    reshape.name,
                    **self.get_attrs(matched_nodes),
                )
            ],
        )

        return reshape.input

    def get_attrs(self, matched_nodes: Iterable[onnx.NodeProto]) -> Dict:
        reshape, transpose, _ = matched_nodes
        permutation = next(
            onnx.helper.get_attribute_value(attr)
            for attr in transpose.attribute
            if attr.name == "perm"
        )
        shape_to_reshape = self.get_value_info_shape(reshape.output[0])
        if permutation == [0, 1, 4, 2, 5, 3]:
            mode = 'CRD'
            blocksize = shape_to_reshape[2]
        elif permutation == [0, 3, 4, 1, 5, 2]:
            mode = 'DCR'
            blocksize = shape_to_reshape[1]
        else:
            assert False, repr(permutation)
        return {'blocksize': blocksize, 'mode': mode}


def _pattern_condition_checker(nodes_to_check: Iterable[onnx.NodeProto]) -> bool:
    _, transpose, _ = nodes_to_check
    perm = next(
        onnx.helper.get_attribute_value(attr) for attr in transpose.attribute if attr.name == "perm"
    )
    return perm in ([0, 1, 4, 2, 5, 3], [0, 3, 4, 1, 5, 2])



from typing import Dict, Iterable, List, Sequence

import numpy as np
import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.interfaces.transformer import Transformer


class ConvertConv1dToConv2d(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [
            Pattern_1,
        ]:
            model = transformer(model).transform()
        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Conv(1d) --> next
    to
        prev --> Reshape --> Conv(2d) --> Reshape --> next

    if Conv(1d).input[0].ndim == 3
    """

    pattern_to_match = ['Conv']

    def pattern_matching(self, base_node: onnx.NodeProto) -> Iterable[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=self.make_new_node(matched_nodes),
            inits_to_add=self.make_new_init(matched_nodes),
            vis_to_add=self.make_new_vi(matched_nodes),
        )
        conv1d = matched_nodes[0]
        return conv1d.input

    def pattern_condition_checker(self, nodes_to_check: Iterable[onnx.NodeProto]) -> bool:
        (conv1d,) = nodes_to_check
        return len(self.get_value_info_shape(conv1d.input[0])) == 3

    def make_new_node(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.NodeProto]:
        (conv1d,) = matched_nodes
        reshape_0 = onnx.helper.make_node(
            'Reshape',
            [conv1d.input[0], conv1d.input[0] + '_reshape'],
            [conv1d.input[0] + '_reshaped'],
            conv1d.name + '_0',
        )
        conv2d = onnx.helper.make_node(
            'Conv',
            [
                reshape_0.output[0],
                conv1d.input[1] + '_expanded',
                conv1d.input[2] if len(conv1d.input) == 3 else "",
            ],
            [conv1d.output[0] + '_expanded'],
            conv1d.name + '_1',
            **_get_conv2d_attrs(conv1d, kernel_size=self.get_value_info_shape(conv1d.input[1])[-1]),
        )
        reshape_1 = onnx.helper.make_node(
            'Reshape',
            [conv2d.output[0], conv1d.output[0] + '_reshape'],
            [conv1d.output[0]],
            conv1d.name + '_2',
        )
        return [reshape_0, conv2d, reshape_1]

    def make_new_init(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.TensorProto]:
        (conv1d,) = matched_nodes
        conv2d_weight_shape = self._get_expanded_shape(conv1d.input[1])
        new_reshape_0 = self._get_expanded_shape(conv1d.input[0])
        new_reshape_1 = self.get_value_info_shape(conv1d.output[0])
        return [
            onnx.numpy_helper.from_array(
                np.array(new_reshape_0), name=conv1d.input[0] + '_reshape'
            ),
            onnx.numpy_helper.from_array(
                self.get_initializer_array(conv1d.input[1]).reshape(conv2d_weight_shape),
                name=conv1d.input[1] + '_expanded',
            ),
            self.initializer_map.get(conv1d.input[2]) if len(conv1d.input) == 3 else None,
            onnx.numpy_helper.from_array(
                np.array(new_reshape_1), name=conv1d.output[0] + '_reshape'
            ),
        ]

    def make_new_vi(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.ValueInfoProto]:
        (conv1d,) = matched_nodes
        conv2d_input_shape = self._get_expanded_shape(conv1d.input[0])
        conv2d_output_shape = self._get_expanded_shape(conv1d.output[0])
        return [
            onnx.helper.make_tensor_value_info(
                conv1d.input[0] + '_reshaped', onnx.TensorProto.FLOAT, conv2d_input_shape
            ),
            onnx.helper.make_tensor_value_info(
                conv1d.output[0] + '_expanded', onnx.TensorProto.FLOAT, conv2d_output_shape
            ),
        ]

    def _get_expanded_shape(self, tensor: str) -> Sequence[int]:
        # Inserts an axis of length one at the position of the H axis.
        # In other words, a tensor of shape (N, C, D) will be reshaped
        # into (N, C, 1, D). We chose the H axis as an insertion point
        # rather than the W axis, i.e. (N, C, D, 1), because the current
        # implemention of our compiler parallelizes computation with
        # tensors with a large W axis better than those with a large H
        # axis. Refer to
        # <https://github.com/furiosa-ai/dss/pull/183#issuecomment-861137438>.
        shape = self.get_value_info_shape(tensor)
        shape.insert(-1, 1)
        return shape


def _get_conv2d_attrs(conv1d: onnx.NodeProto, kernel_size: Sequence[int]) -> Dict:
    attrs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in conv1d.attribute}
    auto_pad = attrs.get('auto_pad', 'NOTSET')
    dilations = attrs.get('dilations', [1])
    group = attrs.get('group', 1)
    kernel_shape = attrs.get('kernel_shape', [kernel_size])
    strides = attrs.get('strides', [1])

    new_attrs = {
        'auto_pad': auto_pad,
        'dilations': [1, *dilations],
        'group': group,
        'kernel_shape': [1, *kernel_shape],
        'strides': [1, *strides],
    }

    if auto_pad == 'NOTSET':
        pads = attrs.get('pads', [0, 0])
        new_attrs.update(
            {
                'pads': [0, pads[0], 0, pads[1]],
            }
        )

    return new_attrs



from typing import Dict, List, Optional

import onnx
import onnxoptimizer

from furiosa.quantizer.frontend.onnx.transformer import utils
from furiosa.quantizer.frontend.onnx.transformer.convert_2d_sum_to_add import Convert2dSumToAdd
from furiosa.quantizer.frontend.onnx.transformer.infer_squeeze_axes import InferSqueezeAxes
from furiosa.quantizer.frontend.onnx.utils.inference_shape import InferenceShape
from furiosa.quantizer.interfaces.transformer import Transformer


class PolishModel(Transformer[onnx.ModelProto]):
    """
    Essential graph transformer/optimizers
    """

    def __init__(self, input_shapes: Optional[Dict[str, List[int]]] = None) -> None:
        super().__init__()
        self.input_shapes = input_shapes

    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        model = utils.name_nodes(model)
        model = utils.make_initializer_name_unique(model)
        model = utils.fix_batch_size_as_one(model)

        model = onnxoptimizer.optimize(model, passes=["extract_constant_to_initializer"])
        model = Convert2dSumToAdd().transform(model)

        model = utils.fixed_point(
            model,
            [
                lambda model: InferenceShape(model).inference_shape(self.input_shapes),
                InferSqueezeAxes().transform,
            ],
        )

        return model



from typing import Dict, Iterable

import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.frontend.onnx.transformer.utils import get_attribute
from furiosa.quantizer.interfaces.transformer import Transformer


class FuseLpNormalization(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [
            Pattern_1,
        ]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> ReduceL2/ReduceL1 --> Clip --> Expand -->  Div --> next
             +                                           +
             -------------------------------------------->
    to
        prev --> LpNormalization --> next
    # TODO Check if Div has no initialzier
    """

    def pattern_matching(self, base_node: onnx.NodeProto) -> Iterable[str]:
        inputs = base_node.input

        pattern_to_match = ['ReduceL2/ReduceL1', 'Clip', 'Expand', 'Div']
        matched_nodes = self.pattern_matcher(base_node, pattern_to_match)
        if not matched_nodes:
            return inputs

        reduce_lp, *_ = matched_nodes
        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=[
                onnx.helper.make_node(
                    'LpNormalization',
                    [reduce_lp.input[0]],
                    [base_node.output[0]],
                    reduce_lp.name,
                    **_get_attrs(reduce_lp),
                )
            ],
        )

        return reduce_lp.input


# TODO Implement Pattern_2 in case of unsimplified graph, containing Shape operator:
# transform
#   prev --> ReduceL2/ReduceL1 --> Clip --> Expand -->  Div --> next
#        +                                +          +
#        ------------------------> Shape ->
#        +                                           +
#        -------------------------------------------->
# to
#   prev --> LpNormalization --> next


def _get_attrs(node: onnx.NodeProto) -> Dict:
    assert node.op_type in ("ReduceL1", "ReduceL2")
    axes = get_attribute(node.attribute, "axes")
    if node.op_type == 'ReduceL1':
        p = 1
    elif node.op_type == 'ReduceL2':
        p = 2
    return {"axis": int(axes[0]), "p": p}



from furiosa.quantizer.frontend.onnx.transformer.experimental.fuse_div_for_bert import (
    FuseDivForBert,
)
from furiosa.quantizer.frontend.onnx.transformer.experimental.reify_conv_for_bert import (
    ReifyConvForBert,
)



import numpy as np
import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.interfaces.transformer import Transformer


class FuseGatherMatMul(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [
            Pattern_1,
        ]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Gather --> MatMul --> next
    to
        prev --> Gather --> next

    if 1. MatMul.ndim == 2
       2. MatMul must have exactly one initializer
    """

    pattern_to_match = ['Gather', 'MatMul']

    def pattern_matching(self, base_node):
        inputs = base_node.input

        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)

        if not matched_nodes:
            return inputs

        if not self.pattern_condition_checker(matched_nodes):
            return inputs

        top_node = matched_nodes[0]

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=[_make_nodes(matched_nodes)],
            inits_to_add=[self.make_initializers(*self.get_new_init_args(matched_nodes))],
        )
        return top_node.input

    def pattern_condition_checker(self, nodes_to_check):
        _, base_node = nodes_to_check

        return self.check_condition_1(base_node.output[0]) and self.check_condition_2(base_node)

    def check_condition_1(self, tensor_name):
        return len(self.get_value_info_shape(tensor_name)) == 2

    def check_condition_2(self, node):
        return sum(node_input in self.initializer_map for node_input in node.input) == 1

    def get_new_init_args(self, matched_nodes):
        top_node, base_node = matched_nodes

        table_input = self.get_init_node_input(top_node)
        matmul_input = self.get_init_node_input(base_node)

        return table_input, matmul_input

    def make_initializers(self, top_node_init, base_node_init):
        table_arr = self.get_initializer_array(top_node_init)
        matmul_arr = self.get_initializer_array(base_node_init)

        new_table_arr = np.matmul(table_arr, matmul_arr)

        new_init = onnx.numpy_helper.from_array(new_table_arr, top_node_init + '_fused')

        return new_init


def _make_nodes(matched_nodes):
    top_node, base_node = matched_nodes

    fused_gather_node = onnx.helper.make_node(
        'Gather',
        inputs=[top_node.input[0] + '_fused', top_node.input[1]],
        outputs=[base_node.output[0]],
        name=base_node.output[0] + '_1',
    )

    return fused_gather_node



# -------------------------------------------------------------------------
#
# FuriosaAI CONFIDENTIAL
# __________________
#
# FuriosaAI Incorporated, All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains
# the property of FuriosaAI and its suppliers,
# if any.  The intellectual and technical concepts contained
# herein are proprietary to FuriosaAI Inc
# and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material
# is strictly forbidden unless prior written permission is obtained
# from FuriosaAI Inc.
# --------------------------------------------------------------------------

from collections import defaultdict
import copy
import itertools
import os
from typing import Dict, List, Optional, Tuple

import numpy as np
import onnx
from onnx import numpy_helper
from onnx.helper import ModelProto, make_node, make_tensor, make_tensor_value_info
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
import tqdm

from furiosa.quantizer.frontend.onnx.quantizer import eliminate_clipper, quantizer_mode
from furiosa.quantizer.frontend.onnx.quantizer.utils import (
    QuantizationMode,
    append_suffix,
    calculate_activation_quant_params,
    calculate_weight_quant_params,
    get_input_tensors,
    is_float_tensor,
)
from furiosa.quantizer.frontend.onnx.transformer import utils
from furiosa.quantizer.frontend.onnx.utils.check_model import check_model

transformers = [
    # eliminate clippers like Relu, Clip into Conv, Add
    # and replace the letters' output quantization parameters with the formers
    # for marginal accuracy gain
    eliminate_clipper.EliminateClipper().transform
]


class FuriosaONNXQuantizer:
    def __init__(
        self,
        model: onnx.ModelProto,
        per_channel: bool,
        static: bool,
        mode: QuantizationMode,
        dynamic_ranges: Dict[str, Tuple[float, float]],
        raw_data=True,
    ):
        """
        - raw_data:
                if True, quantized weight/bias/scale/zero_point/etc.. will be stored in bytes.
                Else, in data_type
        """
        # copy model
        copy_model = ModelProto()
        copy_model.CopyFrom(model)
        self.model = copy_model

        self.raw_data = raw_data

        self.input_tensors = list(map(lambda tensor: tensor[0], get_input_tensors(model)))

        # raise Exception if input_tensor is not defined in model.graph.input
        for input_tensor in self.input_tensors:
            if input_tensor not in [input.name for input in self.model.graph.input]:
                raise Exception(f'input_tensor: {input_tensor} is not defined in model.graph.input')

        # set quantization scheme
        self.per_channel = per_channel
        self.static = static  # TODO develop dynamic quantization scheme?
        self.dynamic_ranges = dynamic_ranges

        # set quantization option guided by mode
        self.mode = mode
        self.activation_qtype = self.weight_qtype = onnx.TensorProto.INT8

        # use uint8 dtype for activation in fake_quant mode
        if mode == QuantizationMode.FAKE:
            self.activation_qtype = onnx.TensorProto.UINT8

        # set model a proto to use
        self.initializer = {init.name: init for init in self.model.graph.initializer}
        self.consumer_map = defaultdict(list)
        for node in model.graph.node:
            for tensor in node.input:
                self.consumer_map[tensor].append(node)
        self.init_vi = [
            make_tensor_value_info(init.name, init.data_type, init.dims)
            for init in self.model.graph.initializer
        ]
        self.value_info_all = {
            vi.name: vi
            for vi in itertools.chain(
                self.model.graph.value_info,
                self.model.graph.input,
                self.model.graph.output,
                self.init_vi,
            )
        }

        # (Case1) check if model is optimized: all value_infos are given with valid dimension/type info.
        utils.check_value_info(self.model)

        # (Case2) raise Exception if dynamic_range is missing
        for key, vi in self.value_info_all.items():
            if not is_float_tensor(vi):
                continue

            if key in self.initializer:
                continue

            if key not in dynamic_ranges:
                raise Exception(f'dynamic_range for {key} is missing')

        # (Case3) raise Exception if dynamic_range is not defined in model.graph.value_info
        for key in dynamic_ranges:
            if key not in self.value_info_all:
                raise Exception(f'dynamic range: {key} is not defined in model.graph.value_info')

        # stack intermediate result of quantization
        self._quant_node = {}
        self._quant_weight = {}
        self._quant_param = {}
        self._quant_value_info = {}

        # quant model.graph field key for quantized model checker
        self._quant_initializer_key = []
        self._quant_value_info_key = []

    def quantize(self) -> onnx.ModelProto:
        # quantize weight and activation
        self.quantize_model()

        # make quantized model and update quantized weight, scale, zero_point
        self.build_quantized_model()

        # check quantized model
        self.check_model()

        return self.model

    def quantize_model(self):
        self._quantize_activation()
        self._quantize_weight()

    def build_quantized_model(self):
        self.model = self.make_intermediate_representation()

        for transform in transformers:
            self.model = transform(self.model)

        if self.mode == QuantizationMode.DFG:
            self.model = quantizer_mode.DFGImportable(self.model, self.raw_data).transform()
        elif self.mode == QuantizationMode.FAKE:
            self.model = quantizer_mode.ONNXRuntimeExecutable(self.model, self.raw_data).transform()

        return self.model

    def make_intermediate_representation(self):
        for node in self.model.graph.node:
            # TODO tmp assumption: Original model with QuantizeLinear and DequantizeLinear is not acceptable.
            if any(op == node.op_type for op in ['QuantizeLinear', 'DequantizeLinear']):
                raise Exception(f'Original model with {node.op_type} is not acceptable.')

            # uses orig_node as an index getter
            # to avoid updating every consumer_map that contains the mutated node
            orig_node = copy.deepcopy(node)
            for idx, node_input in enumerate(node.input):
                if not is_float_tensor(self.value_info_all[node_input]):
                    continue
                if node_input + '_scale' not in self._quant_param:
                    continue

                suffix = (
                    self.consumer_map[node_input].index(orig_node)
                    if len(self.consumer_map[node_input]) > 1
                    else None
                )
                self.make_quant_dequant_node(node_input, suffix)
                node.input[idx] += '_dequantized' + (
                    f'_{str(suffix)}' if suffix is not None else ''
                )

            self._quant_node.update({node.output[0]: node})
        for output in self.model.graph.output:
            if not is_float_tensor(self.value_info_all[output.name]):
                continue
            if output.name + '_scale' not in self._quant_param:
                continue
            self.make_quant_dequant_node(output.name)

            self.model.graph.value_info.append(output)
            output.name += '_dequantized'
            self._quant_value_info.pop(output.name)

        self.model = utils.rebuild_model(
            model=self.model, new_nodes=list(self._quant_node.values()), eliminate=False
        )

        self._update_graph_field(
            field='initializer',
            proto=list(self._quant_param.values()) + list(self.initializer.values()),
        )
        self._update_graph_field(
            field='value_info',
            proto=list(self._quant_value_info.values()) + list(self.model.graph.value_info),
        )

        return self.model

    def make_quant_dequant_node(self, node_input, idx=None):
        scale = node_input + '_scale'
        zero_point = node_input + '_zero_point'
        qlinear_output = node_input + '_quantized'
        dqlinear_output = node_input + '_dequantized'
        if idx is not None:
            dqlinear_output += f'_{str(idx)}'

        # make quantizelinear node
        self._stack_quant_node(
            op_type='QuantizeLinear',
            inputs=[node_input, scale, zero_point],
            outputs=[qlinear_output],
        )
        self._stack_quant_vi_and_qa_helper(
            name=node_input,
            name_quant=qlinear_output,
            elem_type=self._quant_param[zero_point].data_type,
            quant_vi_dict=self._quant_value_info,
        )

        # make dequantizelinear node
        self._stack_quant_node(
            op_type='DequantizeLinear',
            inputs=[qlinear_output, scale, zero_point],
            outputs=[dqlinear_output],
        )
        self._stack_quant_vi_and_qa_helper(
            name=node_input,
            name_quant=dqlinear_output,
            elem_type=onnx.TensorProto.FLOAT,
            quant_vi_dict=self._quant_value_info,
        )

    def check_model(self):
        check_runnable = True
        if self.mode == QuantizationMode.DFG:
            # pass runnable check, as dfg mode does not assume to run on onnxruntime
            check_runnable = False
        check_model(self.model, check_runnable)

        self._quant_value_info_key = [
            vi.name
            for vi in list(self.model.graph.value_info)
            + list(self.model.graph.input)
            + list(self.model.graph.output)
        ]
        self._quant_initializer_key = [init.name for init in self.model.graph.initializer]

        self._check_quant_initializer()
        self._check_quant_value_info()

        if self.mode == QuantizationMode.DFG:
            self._check_quant_param()

    def _quantize_activation(self):
        act_quant_param = calculate_activation_quant_params(
            self.dynamic_ranges, self.model.graph.node, self.activation_qtype, self.value_info_all
        )
        suffix = ['_zero_point', '_scale']
        for name, (zp, s) in act_quant_param.items():
            zp_name, s_name = append_suffix(name=name, suffix=suffix)

            self._stack_quant_param(
                name_zp=zp_name,
                name_scale=s_name,
                data_type_zp=self.activation_qtype,
                dims=zp.shape,
                vals_zp=zp,
                vals_scale=s,
            )

    def _quantize_weight(self):
        disabled = bool(os.environ.get('TQDM_DISABLE'))
        for node in tqdm.tqdm(
            self.model.graph.node, desc='Quantization', disable=disabled, unit=' operators'
        ):
            if node.op_type == 'Conv':
                self._quantize_conv_weight(node, output_channel_axis=0)
            elif node.op_type == 'ConvTranspose':
                self._quantize_conv_weight(node, output_channel_axis=1)
            elif any(
                node.op_type == op for op in ['MatMul', 'Add', 'Mul', 'Div', 'Sub', 'Concat', 'Pow']
            ):
                self._quantize_matmul_weight(node)
            elif node.op_type == 'Clip':
                self._quantize_clip_minmax(node)
            elif node.op_type == 'Pad':
                self._quantize_pad_constant(node)
            else:
                continue

    def _quantize_pad_constant(self, node):
        mode = utils.get_attribute(node.attribute, "mode", b"constant")

        if mode != b'constant':
            return

        try:
            w_init = self.initializer[node.input[2]]
        except IndexError:
            name = f'{node.input[0]}_constant_value'
            node.input.append(name)

            vi = make_tensor_value_info(name=name, elem_type=onnx.TensorProto.FLOAT, shape=[])
            self.model.graph.input.append(vi)
            self.value_info_all.update({name: vi})

            if not self.raw_data:
                w_init = make_tensor(
                    name=name, data_type=onnx.TensorProto.FLOAT, dims=[], vals=[float(0)]
                )
            else:
                w_init = numpy_helper.from_array(np.array(0.0), name=name)

            self.model.graph.initializer.append(w_init)
            self.initializer.update({name: w_init})

        self._quantize_weight_per_layer(w_init)

    def _quantize_clip_minmax(self, node):
        s = numpy_helper.to_array(self._get_quant_param(node.input[0], '_scale'))
        zp = numpy_helper.to_array(self._get_quant_param(node.input[0], '_zero_point'))

        for tensor_name in node.input:
            if tensor_name not in self.initializer:
                continue

            self._stack_quant_param(
                name_zp=tensor_name + '_zero_point',
                name_scale=tensor_name + '_scale',
                data_type_zp=self.activation_qtype,
                dims=zp.shape,
                vals_zp=zp,
                vals_scale=s,
            )

    def _quantize_matmul_weight(self, node):
        for tensor_name in node.input:
            if tensor_name not in self.initializer:
                continue
            w_init = self.initializer[tensor_name]
            self._quantize_weight_per_layer(w_init)

    def _quantize_conv_weight(self, node, output_channel_axis):
        try:
            w_init = self.initializer[node.input[1]]
        except KeyError:
            return

        if self.per_channel:
            self._quantize_weight_per_axis(w_init, axis=output_channel_axis)
        else:
            self._quantize_weight_per_layer(w_init)

        if len(node.input) == 3:
            b_init = self.initializer[node.input[2]]
            i_scale = self._get_quant_param(node.input[0], '_scale')
            w_scale = self._get_quant_param(node.input[1], '_scale')
            self._quantize_bias(b_init, input_scale=i_scale, weight_scale=w_scale)

    def _quantize_weight_per_layer(self, weight_init: onnx.TensorProto) -> None:
        weight = numpy_helper.to_array(weight_init)
        zp, s = calculate_weight_quant_params(
            data=weight.flatten(), weight_qtype=self.weight_qtype, name=weight_init.name
        )

        suffix = ['_quantized', '_zero_point', '_scale']
        _, zp_name, s_name = append_suffix(name=weight_init.name, suffix=suffix)

        self._stack_quant_param(
            name_zp=zp_name,
            name_scale=s_name,
            data_type_zp=self.weight_qtype,
            dims=[],
            vals_zp=np.asarray(zp, dtype=TENSOR_TYPE_TO_NP_TYPE[self.weight_qtype]),
            vals_scale=np.asarray(s, dtype=np.float32),
        )

    def _quantize_weight_per_axis(self, weight_init: onnx.TensorProto, axis: int) -> None:
        assert len(weight_init.dims) == 4, f'weight should have rank 4: {repr(weight_init)}'
        weight = numpy_helper.to_array(weight_init)

        num_output_channels = weight.shape[axis]
        s_list = []
        zp_list = []
        for i in range(num_output_channels):
            indices = [slice(None)] * weight.ndim
            indices[axis] = i
            per_channel_weight = weight[tuple(indices)].flatten()
            zp, s = calculate_weight_quant_params(
                per_channel_weight, self.weight_qtype, weight_init.name
            )
            s_list.append(s)
            zp_list.append(zp)

        suffix = ['_quantized', '_zero_point', '_scale']
        _, zp_name, s_name = append_suffix(name=weight_init.name, suffix=suffix)

        self._stack_quant_param(
            name_zp=zp_name,
            name_scale=s_name,
            data_type_zp=self.weight_qtype,
            dims=[num_output_channels],
            vals_zp=np.asarray(zp_list, dtype=TENSOR_TYPE_TO_NP_TYPE[self.weight_qtype]),
            vals_scale=np.asarray(s_list, dtype=np.float32),
        )

    def _quantize_bias(
        self,
        b_init: onnx.TensorProto,
        input_scale: onnx.TensorProto,
        weight_scale: onnx.TensorProto,
    ) -> None:
        b_scale = numpy_helper.to_array(input_scale) * numpy_helper.to_array(weight_scale)
        b_scale = np.where(b_scale == 0.0, np.float32(2**-149), b_scale)

        qtype = onnx.TensorProto.INT32
        b_zero_point = np.zeros_like(b_scale).astype(TENSOR_TYPE_TO_NP_TYPE[qtype])

        suffix = ['_quantized', '_zero_point', '_scale']
        _, zp_name, s_name = append_suffix(name=b_init.name, suffix=suffix)

        self._stack_quant_param(
            name_zp=zp_name,
            name_scale=s_name,
            data_type_zp=qtype,
            dims=b_zero_point.shape,
            vals_zp=b_zero_point,
            vals_scale=b_scale,
        )

    def _stack_quant_node(
        self,
        inputs: List[str],
        outputs: List[str],
        op_type: str,
        attributes: Optional[List[onnx.AttributeProto]] = None,
    ) -> None:
        if attributes is None:
            attributes = []

        # make quantized node proto
        attr_kwargs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in attributes}

        quant_node = make_node(op_type, inputs, outputs, **attr_kwargs)

        # stack quantized node
        self._quant_node.update({outputs[0]: quant_node})

    def _stack_quant_param(self, name_zp, name_scale, data_type_zp, dims, vals_zp, vals_scale):
        if name_zp in self._quant_param or name_scale in self._quant_param:
            return
        # make quantization parameters initializer proto
        if self.raw_data:
            init_zp = make_tensor(
                name=name_zp, data_type=data_type_zp, dims=dims, vals=vals_zp.tobytes(), raw=True
            )
            init_scale = make_tensor(
                name=name_scale,
                data_type=onnx.TensorProto.FLOAT,
                dims=dims,
                vals=vals_scale.tobytes(),
                raw=True,
            )
        else:
            # Converts numpy 0d array to list of it,
            # for onnx.helper.make_tensor requires vals to be a list.
            vals_zp = np.atleast_1d(vals_zp)
            vals_scale = np.atleast_1d(vals_scale)
            init_zp = make_tensor(name=name_zp, data_type=data_type_zp, dims=dims, vals=vals_zp)
            init_scale = make_tensor(
                name=name_scale, data_type=onnx.TensorProto.FLOAT, dims=dims, vals=vals_scale
            )

        # stack quantization parameters
        self._quant_param.update({init_zp.name: init_zp, init_scale.name: init_scale})

    def _stack_quant_vi_and_qa_helper(self, name, name_quant, elem_type, quant_vi_dict):
        self._stack_quant_value_info(
            name=name, name_quant=name_quant, elem_type=elem_type, quant_vi_dict=quant_vi_dict
        )

        if elem_type == onnx.TensorProto.FLOAT:
            return

    def _stack_quant_value_info(self, name, name_quant, elem_type, quant_vi_dict):
        if name_quant in quant_vi_dict:
            return

        vi = self.value_info_all[name]

        # make quantized value info proto
        quant_vi = make_tensor_value_info(
            name=name_quant,
            elem_type=elem_type,
            shape=[v.dim_value for v in vi.type.tensor_type.shape.dim],
        )

        # stack quantization value info
        quant_vi_dict.update({quant_vi.name: quant_vi})

    def _update_graph_field(self, field, proto):
        self.model.graph.ClearField(field)
        getattr(self.model.graph, field).extend(proto)

    def _check_quant_initializer(self):
        for init in self.model.graph.initializer:
            init_dtype = init.data_type
            # check quant params' data type and field
            if init.name.split('_')[-1] == 'scale':
                assert (
                    init_dtype == onnx.TensorProto.FLOAT
                ), f'Wrong data type {onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[init_dtype]} for {init.name}'
                if not self.raw_data:
                    assert (
                        init.float_data and not init.raw_data
                    ), f'the scale {init.name} should be stored in float_data and not in raw_data: {init}'
                else:
                    assert (
                        init.raw_data and not init.float_data
                    ), f'the scale {init.name} should be stored in raw_data and not in float_data: {init}'

            elif (
                init.name.split('_')[-1] == 'quantized'
                and '_'.join(init.name.split('_')[-2:]) != 'fake_quantized'
            ):
                assert init_dtype in [
                    self.weight_qtype,
                    self.activation_qtype,
                    onnx.TensorProto.INT32,
                ], f'Wrong data type {onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[init_dtype]} for {init.name}'

                if not self.raw_data:
                    assert (
                        init.int32_data and not init.raw_data
                    ), f'the quatized weight/bias {init.name} should be stored in int32_data and not in raw_data: {init}'
                else:
                    assert (
                        init.raw_data and not init.int32_data
                    ), f'the quatized weight/bias {init.name} should be stored in raw_data and not in int32_data: {init}'
            elif any(
                '_'.join(init.name.split('_')[-2:]) == word
                for word in ('zero_point', 'quantized_min', 'quantized_max')
            ):
                assert init_dtype in [
                    self.weight_qtype,
                    self.activation_qtype,
                    onnx.TensorProto.INT32,
                ], f'Wrong data type {onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[init_dtype]} for {init.name}'

                if not self.raw_data:
                    assert (
                        init.int32_data and not init.raw_data
                    ), f'the quatized weight/bias {init.name} should be stored in int32_data and not in in raw_data: {init}'
                else:
                    assert (
                        init.raw_data and not init.int32_data
                    ), f'the quatized weight/bias {init.name} should be stored in raw_data and not in int32_data: {init}'
            elif '_'.join(init.name.split('_')[-2:]) == 'fake_quantized':
                assert (
                    init_dtype == onnx.TensorProto.FLOAT
                ), f'Wrong data type {onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[init_dtype]} for {init.name}'
            else:
                assert init_dtype in (
                    onnx.TensorProto.INT64,
                    onnx.TensorProto.FLOAT,
                ), f'Unknown data type {onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[init_dtype]} for {init.name}'

        # Checks if scale/zero-point of DequantizedLienar/QuantizeLinear (OpSet < 13) are scalars.
        opset = next((opset for opset in self.model.opset_import if not opset.domain), None)
        if opset is not None and opset.version < 13:
            for node in self.model.graph.node:
                if node.op_type == "DequantizeLinear":
                    # Bypasses the checker if node.input[0] in DequantizeLinear is
                    # defined in model.graph.initializer.
                    # Since it might have 1-d scale and zero-point if per-channel quantized,
                    # which conflicts with DequantizeLinear(opset-12) spec
                    # but also is unavoidable according to our graph representations.
                    if node.input[0] in self._quant_initializer_key:
                        continue
                    scale = self._quant_param[node.input[1]]
                    zero_point = self._quant_param[node.input[2]]
                    assert (
                        not scale.dims
                    ), f"the 'x_scale' input of DequantizeLinear (OpSet {opset.version}) must be a scalar"
                    assert (
                        not zero_point.dims
                    ), f"the 'x_zero_point' input of DequantizeLinear (OpSet {opset.version}) must be a scalar"
                elif node.op_type == "QuantizeLinear":
                    scale = self._quant_param[node.input[1]]
                    zero_point = self._quant_param[node.input[2]]
                    assert (
                        not scale.dims
                    ), f"the 'y_scale' input of QuantizeLinear (OpSet {opset.version}) must be a scalar"
                    assert (
                        not zero_point.dims
                    ), f"the 'y_zero_point' input of QuantizeLinear (OpSet {opset.version}) must be a scalar"

        # Checks if quantization parameters are at best 1-d array.
        for init in self.model.graph.initializer:
            postfix = init.name.rsplit('_', maxsplit=1)[-1]
            if postfix not in ['scale', 'zero_point']:
                continue
            quant_param = self._quant_param[init.name]
            rank = len(quant_param.dims)
            assert rank <= 1, f"{init.name} has rank {rank}. {postfix} cannot have rank > 1."

    def _check_quant_value_info(self):
        quant_inputs = [
            node_input
            for node in self.model.graph.node
            for node_input in node.input
            if node_input not in self._quant_initializer_key
        ]
        quant_outputs = [
            node_output for node in self.model.graph.node for node_output in node.output
        ]

        # check if every node.input/output has graph.value_info
        for name in set(quant_inputs + quant_outputs):
            if name not in self._quant_value_info_key:
                raise KeyError(f'{name} is not defined in graph.value_info')

        # check if graph.value_info and graph.input/output are disjoint
        for vi in self.model.graph.value_info:
            for inp in self.model.graph.input:
                if vi.name == inp.name:
                    raise Exception(f'{vi.name} in graph.value_info is also defined in graph.input')

            for oup in self.model.graph.output:
                if vi.name == oup.name:
                    raise Exception(
                        f'{vi.name} in graph.value_info is also defined in graph.output'
                    )

    def _check_quant_param(self):
        # ensure that all scales are non-zero.
        for init in self.model.graph.initializer:
            if init.name.split('_')[-1] == 'scale':
                init_array = onnx.numpy_helper.to_array(init)
                assert all(
                    v != 0.0 for v in np.nditer(init_array)
                ), f"the scale '{init.name}' should be non-zero: {init_array}"

        # check if conv bias scale is correct
        for node in self.model.graph.node:
            if node.op_type != 'QLinearConv':
                continue

            if len(node.input) != 9:
                continue

            i_scale_arr = numpy_helper.to_array(self._quant_param[node.input[1]])
            w_scale_arr = numpy_helper.to_array(self._quant_param[node.input[4]])
            b_scale_name = node.input[-1].split('_quantized')[0] + '_scale'
            b_scale_arr = numpy_helper.to_array(self._quant_param[b_scale_name])

            assert np.allclose(
                b_scale_arr, (i_scale_arr * w_scale_arr).reshape(-1)
            ), f'Conv bias scale is incorrect: {b_scale_name}'

        # check if transposeconv bias scale is correct
        for node in self.model.graph.node:
            if node.op_type != "ConvTranspose":
                continue

            if len(node.input) != 3:
                continue

            i_scale_arr = numpy_helper.to_array(
                self._quant_param[node.input[0].split('_dequantized')[0] + '_scale']
            )
            w_scale_arr = numpy_helper.to_array(
                self._quant_param[node.input[1].split('_dequantized')[0] + '_scale']
            )
            b_scale_name = node.input[2].split('_dequantized')[0] + '_scale'
            b_scale_arr = numpy_helper.to_array(self._quant_param[b_scale_name])

            assert np.allclose(
                b_scale_arr, (i_scale_arr * w_scale_arr).reshape(-1)
            ), f'Conv bias scale is incorrect: {b_scale_name}'

    def _get_quant_param(self, origin, postfix=None):
        result = self._quant_param.get(f'{origin}{postfix or ""}', None)
        if result is None:
            raise Exception(f"dynamic-range '{origin}' is missing")
        return result



from typing import Iterable, List

import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.interfaces.transformer import Transformer


# TODO change name to express general objective of transformer.
class InferSqueezeAxes(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [
            Pattern_1,
        ]:
            model = transformer(model).transform()

        return model


# TODO case when model's opset > 12 and Squeeze's input axes does not exist
class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Squeeze (axes attribute is None) --> next
    to
        prev --> Squeeze (axes attribute is filled using input's value info) --> next

    if  1. model's opset < 13.
        2. axes attribute of Squeeze does not exist
        3. Squeeze.input[0] has shape info (graph input or shape inferred value info)
    """

    pattern_to_match = ['Squeeze']

    def pattern_matching(self, base_node: onnx.NodeProto) -> List[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=self.make_new_node(matched_nodes),
        )

        return base_node.input

    def pattern_condition_checker(self, nodes_to_check: Iterable[onnx.NodeProto]) -> bool:
        opset = next((opset for opset in self.model.opset_import if not opset.domain), None)
        if opset is None or opset.version >= 13:
            return False
        (squeeze,) = nodes_to_check
        try:
            if not self.get_value_info_shape(squeeze.input[0]):
                return False
        except ValueError:
            return False
        return all(attr.name != "axes" for attr in squeeze.attribute)

    def make_new_node(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.NodeProto]:
        (squeeze,) = matched_nodes
        input_shape = self.get_value_info_shape(squeeze.input[0])
        axes = [i for (i, dim_value) in enumerate(input_shape) if dim_value == 1]

        new_squeeze = onnx.helper.make_node(
            'Squeeze',
            inputs=squeeze.input,
            outputs=squeeze.output,
            name=squeeze.name,
            axes=axes,
        )
        return [new_squeeze]



import onnx
from onnx import numpy_helper
from onnx.helper import make_model

from furiosa.quantizer.frontend.onnx.transformer import utils
from furiosa.quantizer.frontend.onnx.transformer.polish_model import PolishModel
from furiosa.quantizer.frontend.onnx.utils.check_model import check_model
from furiosa.quantizer.interfaces.transformer import Transformer


class FuseDivForBert(Transformer):
    """
    Only works for some BERT Models
    """

    def __init__(self):
        self.nodes_by_output_name = None
        self.initializers = None
        self.outputs_by_name = None

    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        model = PolishModel().transform(model)

        self.nodes_by_output_name = {node.output[0]: node for node in model.graph.node}
        self.initializers = {init.name: init for init in model.graph.initializer}
        self.outputs_by_name = {oup.name: oup for oup in model.graph.output}

        model = self.transform_matmul_add(model)  # transform matmul + add --> conv
        check_model(model)

        return PolishModel().transform(model)

    def transform_matmul_add(self, model):
        optimized_nodes = []
        removed_nodes = []

        # Handle Case1: MatMul + Add
        for node in model.graph.node:
            if node.op_type != 'Div':
                optimized_nodes.append(node)
                continue

            prev_node_1 = self.nodes_by_output_name[node.input[0]]
            if prev_node_1.op_type != 'MatMul':
                optimized_nodes.append(node)
                continue

            prev_node_2 = self.nodes_by_output_name[prev_node_1.input[0]]
            prev_node_3 = self.nodes_by_output_name[prev_node_2.input[0]]
            prev_node_4 = self.nodes_by_output_name[prev_node_3.input[0]]

            assert prev_node_4.op_type == 'Add'

            scalar = numpy_helper.to_array(self.initializers[node.input[1]])
            arr = numpy_helper.to_array(self.initializers[prev_node_4.input[1]])

            model.graph.initializer.append(
                numpy_helper.from_array(arr / scalar, name=prev_node_4.input[1] + '_div_fused')
            )
            prev_node_4.input[1] = prev_node_4.input[1] + '_div_fused'
            prev_node_1.output[0] = node.output[0]
            removed_nodes.append(node)

        new_nodes = list(filter(lambda node: node not in removed_nodes, optimized_nodes))
        model.graph.ClearField('node')
        model.graph.node.extend(new_nodes)
        model = make_model(model.graph)

        model = utils.eliminate_unused_protos(model)

        return model



import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.frontend.onnx.transformer.utils import get_attribute
from furiosa.quantizer.interfaces.transformer import Transformer


class FusePad(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [Pattern_1, Pattern_2]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Pad --> MaxPool --> next
    to
        prev --> MaxPool --> next

    if 1. Pad.mode == 'constant'
       2. Pad.constant_value == -inf
       3. padded on spatial dimension
       4. fused_pads[i] < kernel_shape[i] and fused_pads[i + kernel_rank] < kernel_shape[i] for all i
    """

    pattern_to_match = ['Pad', 'MaxPool']

    def pattern_matching(self, base_node):
        inputs = base_node.input

        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return inputs

        if not self.pattern_condition_checker(matched_nodes):
            return inputs

        self.transform_to_fuse(matched_nodes, nodes_to_add=[self.make_new_node(matched_nodes)])

        return matched_nodes[0].input

    def pattern_condition_checker(self, nodes_to_check):
        top_node, base_node = nodes_to_check

        return (
            _get_pad_mode(top_node) == b'constant'
            and self.check_condition_2(top_node)
            and self.check_condition_3(top_node.input[1])
            and self.check_condition_6(self.get_attrs(base_node), top_node.input[1])
        )

    def check_condition_2(self, node):
        try:
            const_input = node.input[2]
            constant_value = self.get_initializer_array(const_input)
        except IndexError:
            constant_value = 0.0
        return constant_value == float('-inf')

    def check_condition_3(self, pads_input):
        pads = self.get_initializer_array(pads_input)
        rank = len(pads) // 2
        pads_on_nc_dim = [*pads[:2], *pads[rank : rank + 2]]
        return all(pad == 0 for pad in pads_on_nc_dim)

    def check_condition_6(self, node_attrs, pad_input):
        attrs = self.update_attrs(node_attrs, pad_input)
        kernel_shape = attrs['kernel_shape']
        kernel_rank = len(kernel_shape)
        fused_pads = attrs['pads']
        fused_pad_shape = [
            sum([fused_pads[dim], fused_pads[dim + kernel_rank]]) for dim in range(kernel_rank)
        ]

        assert len(kernel_shape) == len(fused_pad_shape)
        return all(
            fused_pads[dim] < k and fused_pads[dim + kernel_rank] < k
            for dim, k in enumerate(kernel_shape)
        )

    def update_attrs(self, attrs, pad_input):
        pads = [sum(x) for x in zip(attrs['pads'], self.make_maxpool_pad(pad_input))]
        attrs['pads'] = pads

        return attrs

    def get_attrs(self, node):
        rank = len(self.get_value_info_shape(node.input[0]))
        nspatial_dim = rank - 2

        attrs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in node.attribute}
        ceil_mode = attrs.get('ceil_mode', 0)
        dilations = attrs.get('dilations', [1] * nspatial_dim)
        kernel_shape = attrs['kernel_shape']
        strides = attrs.get('strides', [1] * nspatial_dim)
        pads = attrs.get('pads', [0] * nspatial_dim * 2)

        return {
            'ceil_mode': ceil_mode,
            'dilations': dilations,
            'kernel_shape': kernel_shape,
            'pads': pads,
            'strides': strides,
        }

    def make_maxpool_pad(self, pad_input):
        pads = self.get_initializer_array(pad_input)
        rank = len(pads) // 2

        new_pads = []
        for pad in pads:
            if pad == -1:
                new_pads.append(0)
            else:
                new_pads.append(pad)
        pads = new_pads

        return [*pads[2:rank], *pads[rank + 2 : 2 * rank]]

    def make_new_node(self, matched_nodes):
        top_node, base_node = matched_nodes
        attrs = self.update_attrs(self.get_attrs(base_node), top_node.input[1])

        return onnx.helper.make_node(
            'MaxPool', [top_node.input[0]], [base_node.output[0]], name=top_node.name, **attrs
        )


class Pattern_2(Pattern_1):
    """
    transform
        prev --> Pad --> AveragePool --> next
    to
        prev --> AveragePool --> next

    if 1. Pad.mode == 'constant'
       2. Pad.constant_value == 0.0
       3. padded on spatial dimension
       4. AveragePool.count_include_pad == 1 or all AveragePool.pads == 0
       5. AveragePool.ceil_mode == 0
       6. fused_pads[i] < kernel_shape[i] and fused_pads[i + kernel_rank] < kernel_shape[i] for all i
    """

    pattern_to_match = ['Pad', 'AveragePool']

    def pattern_condition_checker(self, nodes_to_check):
        top_node, base_node = nodes_to_check
        return (
            _get_pad_mode(top_node) == b'constant'
            and self.check_condition_2(top_node)
            and self.check_condition_3(top_node.input[1])
            and self.check_condition_4(base_node)
            and self.check_condition_5(base_node)
            and self.check_condition_6(self.get_attrs(base_node), top_node.input[1])
        )

    def get_attrs(self, node):
        rank = len(self.get_value_info_shape(node.input[0]))
        nspatial_dim = rank - 2

        attrs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in node.attribute}
        ceil_mode = attrs.get('ceil_mode', 0)
        count_include_pad = attrs.get('count_include_pad', 0)
        kernel_shape = attrs['kernel_shape']
        strides = attrs.get('strides', [1] * nspatial_dim)
        pads = attrs.get('pads', [0] * nspatial_dim * 2)

        return {
            'ceil_mode': ceil_mode,
            'count_include_pad': count_include_pad,
            'kernel_shape': kernel_shape,
            'pads': pads,
            'strides': strides,
        }

    def update_attrs(self, attrs, pad_input):
        pads = [sum(x) for x in zip(attrs['pads'], self.make_maxpool_pad(pad_input))]
        attrs['pads'] = pads
        attrs['count_include_pad'] = 1

        return attrs

    def check_condition_2(self, node):
        try:
            const_input = node.input[2]
            constant_value = self.get_initializer_array(const_input)
        except IndexError:
            constant_value = 0.0

        if constant_value == 0.0:
            return True
        return False

    def check_condition_4(self, node):
        attrs = self.get_attrs(node)
        count_include_pad = attrs['count_include_pad']
        pads = attrs['pads']
        if count_include_pad == 1 or all(pad == 0 for pad in pads):
            return True
        return False

    def check_condition_5(self, node):
        attrs = self.get_attrs(node)
        ceil_mode = attrs['ceil_mode']
        if ceil_mode == 0:
            return True
        return False

    def make_new_node(self, matched_nodes):
        top_node, base_node = matched_nodes
        attrs = self.update_attrs(self.get_attrs(base_node), top_node.input[1])

        return onnx.helper.make_node(
            'AveragePool', [top_node.input[0]], [base_node.output[0]], name=top_node.name, **attrs
        )


def _get_pad_mode(node: onnx.NodeProto):
    return get_attribute(node.attribute, "mode", b"constant")



import os
import tempfile
from typing import Dict, Iterable, Iterator, Optional, Tuple

import numpy as np
import onnx
import onnxruntime.quantization.calibrate
import tqdm


class CalibrationError(Exception):
    """The base class for all exceptions that are related to calibration."""


class CalibrationDataReaderForIterator(onnxruntime.quantization.calibrate.CalibrationDataReader):
    """A CalibrationDataReader that wraps dicts mapping input tensor names to their values."""

    def __init__(self, iterator: Iterator[Dict[str, np.ndarray]]):
        self.iterator = iterator

    def get_next(self):
        return next(self.iterator, None)


def calibrate(
    model: onnx.ModelProto,
    dataset: Iterable[Dict[str, np.ndarray]],
    augmented_model_path: Optional[str] = None,
) -> Dict[str, Tuple[float, float]]:
    """Estimates the range of tensors in a model, based on a dataset.

    Args:
        model: An ONNX model to calibrate.
        dataset: An Iterable that returns dicts mapping input tensor names to their values.
        augmented_model_path: A path to save an augmented model to.

    Returns:
        A dict mapping tensors in the model to their minimum and maximum values.
    """
    if augmented_model_path is None:
        with tempfile.NamedTemporaryFile(suffix=".onnx", delete=False) as f:
            augmented_model_path = f.name

    calibrator = onnxruntime.quantization.calibrate.MinMaxCalibrater(
        model, augmented_model_path=augmented_model_path
    )
    if os.environ.get("TQDM_DISABLE"):
        dataset = iter(dataset)
    else:
        dataset = tqdm.tqdm(dataset, desc="Calibration", unit=" samples")
    calibrator.collect_data(CalibrationDataReaderForIterator(iter(dataset)))
    return calibrator.compute_range()


def calibrate_with_random_data(
    model: onnx.ModelProto, dataset_size: int = 8, augmented_model_path: Optional[str] = None
) -> Dict[str, Tuple[float, float]]:
    """Estimates the range of tensors in a model, based on a random dataset.

    Args:
        model: An ONNX model to calibrate.
        dataset_size: the size of a random dataset to use.
        augmented_model_path: A path to save an augmented model to.

    Returns:
        A dict mapping tensors in the model to their minimum and maximum values.
    """
    if augmented_model_path is None:
        with tempfile.NamedTemporaryFile(suffix=".onnx", delete=False) as f:
            augmented_model_path = f.name

    calibrator = onnxruntime.quantization.calibrate.MinMaxCalibrater(
        model, augmented_model_path=augmented_model_path
    )
    initializers = set(tensor.name for tensor in model.graph.initializer)
    rng = np.random.default_rng()
    for _ in range(dataset_size):
        inputs = {}
        for value_info in model.graph.input:
            if value_info.name in initializers:
                continue
            # https://github.com/onnx/onnx/blob/master/docs/IR.md#static-tensor-shapes
            #
            # > The static shape is defined by 'TensorShapeProto':
            # >
            # >     message TensorShapeProto {
            # >       message Dimension {
            # >         oneof value {
            # >           int64 dim_value = 1;
            # >           string dim_param = 2;
            # >         };
            # >       };
            # >       repeated Dimension dim = 1;
            # >     }
            # >
            # > Which is referenced by the Tensor type message:
            # >
            # >     message Tensor {
            # >       optional TensorProto.DataType elem_type = 1;
            # >       optional TensorShapeProto shape = 2;
            # >     }
            # >
            # > The empty list of dimension sizes, [], is a valid tensor shape, denoting a
            # > zero-dimension (scalar) value. A zero-dimension tensor is distinct from a tensor of
            # > unknown dimensionality, which is indicated by an absent 'shape' property in the
            # > Tensor message. When the shape property is absent in the type of a value (including
            # > node input), it indicates that the corresponding runtime value may have any shape.
            # > This sub-section describes how to interpret a missing-shape or a shape with missing
            # > dimensions etc. However, specific usage contexts may impose further constraints on a
            # > type and shape. For example, the inputs and outputs of a model (top-level graph) are
            # > required to have a shape, indicating the rank of inputs and outputs, even though the
            # > exact dimensions need not be specified.
            shape = []
            for dimension in value_info.type.tensor_type.shape.dim:
                if dimension.HasField("dim_value"):
                    shape.append(dimension.dim_value)
                else:
                    raise CalibrationError(
                        f"The static shape of tensor '{value_info.name}' must be provided"
                    )
            if value_info.type.tensor_type.elem_type == onnx.TensorProto.DataType.FLOAT:
                inputs[value_info.name] = rng.standard_normal(size=shape, dtype=np.float32)
            else:
                raise NotImplementedError(
                    onnx.TensorProto.DataType.Name(value_info.type.tensor_type.elem_type)
                )
        calibrator.collect_data(CalibrationDataReaderForIterator(iter([inputs])))
    return calibrator.compute_range()



import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.interfaces.transformer import Transformer


class FuseRedundantReshapePattern(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [
            Pattern_2,
            Pattern_1,
            Pattern_3,
        ]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Reshape --> Reshape --> next
    to
        prev --> Reshape --> next

    if prev.output[0].shape != next.input[0].shape
    """

    pattern_to_match = ['Reshape', 'Reshape']
    postfix = '_reshape_fused'

    def pattern_matching(self, base_node):
        inputs = base_node.input

        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return inputs

        if not self.pattern_condition_checker(matched_nodes):
            return inputs

        top_node = matched_nodes[0]

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=[self.make_new_node(matched_nodes)],
            inits_to_add=[self.make_new_init(matched_nodes)],
        )
        return top_node.input

    def pattern_condition_checker(self, nodes_to_check):
        top_node = nodes_to_check[0]
        base_node = nodes_to_check[-1]
        if self.is_same_shape(top_node.input[0], base_node.output[0]):
            return False
        return True

    def make_new_node(self, matched_nodes):
        top_node = matched_nodes[0]
        base_node = matched_nodes[-1]
        return onnx.helper.make_node(
            'Reshape',
            [top_node.input[0], top_node.input[1] + self.postfix],
            [base_node.output[0]],
            name=top_node.name,
        )

    def make_new_init(self, matched_nodes):
        top_node = matched_nodes[0]
        base_node = matched_nodes[-1]
        return self.make_int64_initializer(top_node.input[1] + self.postfix, base_node.output[0])

    def make_new_vi(self, matched_nodes):
        top_node = matched_nodes[0]
        return self.copy_value_info(top_node.input[0])


class Pattern_2(Pattern_1):
    """
    transform
        prev --> Reshape --> Reshape --> Reshape --> next
    to
        prev --> Reshape --> next

    if prev.output[0].shape != next.input[0].shape
    """

    pattern_to_match = ['Reshape', 'Reshape', 'Reshape']


class Pattern_3(Pattern_1):
    """
    transform
        prev --> Flatten/Squeeze --> Unsqueeze --> next
    to
        prev --> Reshape --> next
    if prev.output[0].shape != next.input[0].shape
    """

    pattern_to_match = ['Flatten/Squeeze', 'Unsqueeze']

    def make_new_node(self, matched_nodes):
        top_node = matched_nodes[0]
        base_node = matched_nodes[-1]
        return onnx.helper.make_node(
            'Reshape',
            [top_node.input[0], top_node.input[0] + self.postfix],
            [base_node.output[0]],
            top_node.name,
        )

    def make_new_init(self, matched_nodes):
        top_node = matched_nodes[0]
        base_node = matched_nodes[-1]
        return self.make_int64_initializer(top_node.input[0] + self.postfix, base_node.output[0])



import onnx
from onnx import numpy_helper
from onnx.helper import make_model, make_node, make_tensor_value_info

from furiosa.quantizer.frontend.onnx.transformer import utils
from furiosa.quantizer.frontend.onnx.transformer.polish_model import PolishModel
from furiosa.quantizer.frontend.onnx.utils.check_model import check_model
from furiosa.quantizer.interfaces.transformer import Transformer


class ReifyConvForBert(Transformer):
    """
    from: MatMul + Add
    to: Conv

    Assume NCHW Input
    """

    def __init__(self):
        self.nodes_by_output_name = None
        self.initializers = None
        self.outputs_by_name = None

    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        model = PolishModel().transform(model)

        self.nodes_by_output_name = {node.output[0]: node for node in model.graph.node}
        self.initializers = {init.name: init for init in model.graph.initializer}
        self.outputs_by_name = {oup.name: oup for oup in model.graph.output}

        model = self.transform_matmul_add(model)  # transform matmul + add --> conv
        check_model(model)

        return PolishModel().transform(model)

    def transform_matmul_add(self, model):
        optimized_nodes = []
        removed_nodes = []

        # Handle Case1: MatMul + Add
        for node in model.graph.node:
            if node.op_type != 'Add':
                optimized_nodes.append(node)
                continue

            # Add has no specific order of input according to spec.
            # Therefore, we need to find the input index of MatMul
            try:
                idx_matmul = [
                    i
                    for i, tensor_name in enumerate(node.input)
                    if tensor_name not in self.initializers
                    and self.nodes_by_output_name[tensor_name].op_type == 'MatMul'
                ]
            except KeyError:
                optimized_nodes.append(node)
                continue

            # Expect one of the inputs is Exp and the other is ReduceSum
            if len(idx_matmul) != 1:
                optimized_nodes.append(node)
                continue

            idx_matmul = idx_matmul[0]
            matmul_node = self.nodes_by_output_name[node.input[idx_matmul]]

            idx_matmul_init = next(
                i
                for i, tensor_name in enumerate(matmul_node.input)
                if tensor_name in self.initializers
            )

            matmul_init = self.initializers[matmul_node.input[idx_matmul_init]]
            matmul_weight = numpy_helper.to_array(matmul_init)
            c, n = matmul_weight.shape
            conv_weight = matmul_weight.transpose().reshape(n, c, 1, 1)

            idx_add_init = next(
                i
                for i, tensor_name in enumerate(matmul_node.input)
                if tensor_name in self.initializers
            )

            add_init = self.initializers[node.input[idx_add_init]]
            bias = numpy_helper.to_array(add_init)

            conv_weight_init = numpy_helper.from_array(
                conv_weight, name=matmul_node.input[idx_matmul_init] + '_reified'
            )
            bias_init = numpy_helper.from_array(bias, name=node.input[idx_add_init] + '_reified')
            model.graph.initializer.extend([conv_weight_init, bias_init])

            removed_nodes.extend([node, matmul_node])

            unsqueeze_node = make_node(
                'Unsqueeze',
                inputs=[matmul_node.input[0]],
                outputs=[matmul_node.output[0] + '_expanded'],
                **{
                    'axes': [
                        1,
                    ]
                },
            )

            transpose_node = make_node(
                'Transpose',
                inputs=[unsqueeze_node.output[0]],
                outputs=[matmul_node.output[0] + '_transposed'],
                **{'perm': [0, 3, 1, 2]},
            )

            conv_node = make_node(
                'Conv',
                inputs=[transpose_node.output[0], conv_weight_init.name, bias_init.name],
                outputs=[matmul_node.output[0] + '_conv_output'],
                **{
                    'dilations': [1, 1],
                    'group': 1,
                    'kernel_shape': [1, 1],
                    'pads': [0, 0, 0, 0],
                    'strides': [1, 1],
                },
            )

            squeeze_node = make_node(
                'Squeeze',
                inputs=[conv_node.output[0]],
                outputs=[matmul_node.output[0] + '_squeezed'],
                **{'axes': [2]},
            )

            transpose_node_1 = make_node(
                'Transpose',
                inputs=[squeeze_node.output[0]],
                outputs=[node.output[0]],
                **{'perm': [0, 2, 1]},
            )

            optimized_nodes.extend(
                [unsqueeze_node, transpose_node, conv_node, squeeze_node, transpose_node_1]
            )

            graph_input = model.graph.input[0]
            if conv_node.input[0] == graph_input.name:
                batch_size = graph_input.type.tensor_type.shape.dim[0].dim_value
                new_vi = make_tensor_value_info(
                    name=graph_input.name,
                    elem_type=graph_input.type.tensor_type.elem_type,
                    shape=(batch_size, c, 1, 1),
                )
                model.graph.input.remove(graph_input)
                model.graph.input.insert(0, new_vi)

        new_nodes = list(filter(lambda node: node not in removed_nodes, optimized_nodes))
        model.graph.ClearField('node')
        model.graph.node.extend(new_nodes)
        model = make_model(model.graph)

        model = utils.eliminate_unused_protos(model)

        return model



from enum import Enum
import logging
from typing import Dict, List, Tuple

import numpy as np
import onnx
from onnx import TensorProto
import onnxruntime as ort

logger = logging.getLogger('Furiosa-Quantizer')
logging.basicConfig(level=logging.INFO)

__PRODUCER__ = "jason_furiosa"


class QuantizationMode(Enum):
    # DFG: Quantize graph to DFG(Quantized graph) export
    DFG = 0
    # FAKE: Evaluate quantized graph replacing QConvLinear with Conv2d/MatMul &
    FAKE = 1


def get_qrange(qtype):
    """
    source: onnxruntime quantization tools
    """
    if qtype == TensorProto.UINT8:
        return 255  # 2^b - 1
    if qtype == TensorProto.INT8:
        return 254  # [-(2^{b-1}-1), 2^{b-1}-1]: [-127, 127] for 8 bits.
    raise ValueError('unsupported quantization data type')


def get_vi_dtype(vi):
    """
    This function returns value_info's data type

    :param vi: graph.value_info
    :return: graph.value_info.type.tensor_type.elem_type
    """
    return vi.type.tensor_type.elem_type


def is_float_tensor(vi):
    if get_vi_dtype(vi) == onnx.TensorProto.FLOAT:
        return True
    return False


def activation_scale_zeropoint(rmin, rmax, activation_qtype):
    rmin = min(rmin, 0)
    rmax = max(rmax, 0)

    return asymmetric_scale_zeropoint(rmin, rmax, activation_qtype)


def asymmetric_scale_zeropoint(rmin, rmax, activation_qtype):
    """
    source: onnxruntime quantization tools
    """
    scale = np.float32((rmax - rmin) / 255 if rmin != rmax else 1)
    # The minimum positive (subnormal) value is 2 ** -149 for IEEE 754 single-precision binary floating-point format
    # source: https://en.wikipedia.org/wiki/Single-precision_floating-point_format#Exponent_encoding
    scale = max(scale, 2**-149)
    if activation_qtype == TensorProto.UINT8:
        initial_zero_point = (0 - rmin) / scale
        zero_point = np.uint8(round(max(0, min(255, initial_zero_point))))
        return np.array(zero_point), np.array(scale)
    if activation_qtype == TensorProto.INT8:
        initial_zero_point = -128 - rmin / scale
        zero_point = np.int8(round(max(-128, min(127, initial_zero_point))))
        return np.array(zero_point), np.array(scale)
    raise Exception('qType must be one of UINT8 or INT8')


def calculate_activation_quant_params(
    dynamic_ranges: Dict,
    node_list: List[onnx.NodeProto],
    activation_qtype: TensorProto,
    value_info: Dict,
) -> Dict:
    quantization_params = {}
    for node in node_list:
        # Quantize activation input/output, following TFLite's quantization specification
        if node.op_type in [
            'MaxPool',
            'Squeeze',
            'Unsqueeze',
            'Gather',
            'Transpose',
            'Reshape',
            'DepthToSpace',
            'Expand',
            'Flatten',
        ]:
            if not is_float_tensor(value_info[node.input[0]]):
                continue
            if node.input[0] not in quantization_params:
                quantization_params[node.input[0]] = activation_scale_zeropoint(
                    *dynamic_ranges[node.input[0]], activation_qtype
                )

            quantization_params[node.output[0]] = quantization_params[node.input[0]]
        elif node.op_type in ['Softmax', 'Sigmoid']:
            if node.input[0] not in quantization_params:
                quantization_params[node.input[0]] = activation_scale_zeropoint(
                    *dynamic_ranges[node.input[0]], activation_qtype
                )

            if activation_qtype == TensorProto.INT8:
                zero_point = np.array((np.int8(-128)))
            elif activation_qtype == TensorProto.UINT8:
                zero_point = np.array(np.uint8(0))
            else:
                raise Exception()
            quantization_params[node.output[0]] = (zero_point, np.array(np.float32(1.0 / 256.0)))
        elif node.op_type in ['LpNormalization']:
            if node.input[0] not in quantization_params:
                quantization_params[node.input[0]] = activation_scale_zeropoint(
                    *dynamic_ranges[node.input[0]], activation_qtype
                )

            if activation_qtype == TensorProto.INT8:
                zero_point = np.array((np.int8(0)))
            elif activation_qtype == TensorProto.UINT8:
                zero_point = np.array(np.uint8(128))
            else:
                raise Exception()
            quantization_params[node.output[0]] = (zero_point, np.array(np.float32(1.0 / 128.0)))
        else:
            for name in list(node.input) + list(node.output):
                if name not in dynamic_ranges:
                    continue
                if name in quantization_params:
                    continue
                rmin, rmax = dynamic_ranges[name]
                zero_point, scale = activation_scale_zeropoint(rmin, rmax, activation_qtype)
                quantization_params[name] = (zero_point, scale)

    return quantization_params


def calculate_weight_quant_params(
    data: np.ndarray, weight_qtype: TensorProto, name: str
) -> Tuple[int, float]:
    """
    :parameter data: data to quantize
    :parameter weight_qtype: quantization data type of weight
    :parameter name: name of tensor to quantize
    :return: quantized weights, zero point, scale

    To pack weights, we compute a linear transformation
        - when data type == uint8 mode, from [rmin, rmax] -> [0, 2^{b-1}] and
        - when data type == int8, from [-m , m] -> [-(2^{b-1}-1), 2^{b-1}-1] where
            m = max(abs(rmin), abs(rmax))

    and add necessary intermediate nodes to trasnform quantized weight to full weight using the equation
    r = S(q-z), where
        r: real original value
        q: quantized value
        S: scale
        z: zero point

    source: onnxruntime quantization tools
    """
    rmin = min(min(data), 0)
    rmax = max(max(data), 0)

    quantized_range = get_qrange(weight_qtype)

    if weight_qtype == TensorProto.INT8:
        max_range = max(abs(rmin), abs(rmax))
        if max_range > 0:
            scale = (max_range * 2.0) / quantized_range
        else:
            logger.info('Both the min and the max of data are 0: %s', name)
            scale = 1.0
        zero_point = 0
    elif weight_qtype == TensorProto.UINT8:
        scale = (float(rmax) - rmin) / quantized_range if rmin != rmax else 1
        zero_point = round((0 - rmin) / scale)  # round to nearest integer
    else:
        raise ValueError(
            f"Unexpected data type {weight_qtype} requested. Only INT8 and UINT8 are supported."
        )

    # The minimum positive (subnormal) value is 2 ** -149 for IEEE 754 single-precision binary floating-point format
    # source: https://en.wikipedia.org/wiki/Single-precision_floating-point_format#Exponent_encoding
    scale = max(scale, 2**-149)
    return zero_point, scale


def append_suffix(name: str, suffix: List[str]) -> List[str]:
    """
    Helper function to append suffixes to the given name.
    """
    return list(map(lambda x: name + x, suffix))


def get_input_tensors(model: onnx.ModelProto) -> List[Tuple[str, List[int], str]]:
    sess = ort.InferenceSession(model.SerializeToString())
    input_tensors = [(inp.name, inp.shape, inp.type) for inp in sess.get_inputs()]
    return input_tensors



import logging
from typing import Callable, Iterable, List, Tuple

import numpy as np
import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer, utils
from furiosa.quantizer.interfaces.transformer import Transformer

logger = logging.getLogger(__name__)


class FuseBatchNorm(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [Pattern_1, Pattern_2, Pattern_3, Pattern_4]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Conv --> BatchNormalization --> next
    to
        prev --> Conv --> next
    """

    pattern_to_match = ['Conv', 'BatchNormalization']

    def pattern_matching(self, base_node: onnx.NodeProto) -> List[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        conv, batch_norm = matched_nodes
        multiplier, shifter = _get_multiplier_and_shifter(
            *_get_bn_params(batch_norm, self.get_initializer_array)
        )

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=_make_bn_fused_node(conv, batch_norm.output[0]),
            inits_to_add=_make_bn_fused_init(conv, multiplier, shifter, self.get_initializer_array),
            vis_to_add=[],
        )

        return conv.input


class Pattern_2(ONNXTransformer):
    """
    transform
        prev --> ConvTranspose --> BatchNormalization --> next
    to
        prev --> ConvTranspose --> next
    """

    pattern_to_match = ['ConvTranspose', 'BatchNormalization']

    def pattern_matching(self, base_node: onnx.NodeProto) -> List[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        conv_trans, batch_norm = matched_nodes
        multiplier, shifter = _get_multiplier_and_shifter(
            *_get_bn_params(batch_norm, self.get_initializer_array)
        )

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=_make_bn_fused_node(conv_trans, batch_norm.output[0]),
            inits_to_add=_make_bn_fused_init(
                conv_trans, multiplier, shifter, self.get_initializer_array
            ),
            vis_to_add=[],
        )

        return conv_trans.input


class Pattern_3(ONNXTransformer):
    """
    transform
        prev --> Conv --> Mul --> Add --> next
    to
        prev --> Conv --> next

    if 1. Mul has only one initializer
       2. Add has only one initializer
    """

    pattern_to_match = ['Conv', 'Mul', 'Add']

    def pattern_matching(self, base_node: onnx.NodeProto) -> List[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        conv, mul, add = matched_nodes
        multiplier = self.get_initializer_array(self.get_init_node_input(mul)).flatten()
        shifter = self.get_initializer_array(self.get_init_node_input(add)).flatten()

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=_make_bn_fused_node(conv, add.output[0]),
            inits_to_add=_make_bn_fused_init(conv, multiplier, shifter, self.get_initializer_array),
            vis_to_add=[],
        )

        return conv.input

    def pattern_condition_checker(self, nodes_to_check: List[onnx.NodeProto]) -> bool:
        _, mul, add = nodes_to_check

        # This checks if a node has a initializer, \
        # assuming a node has exactly one initializer if it has one. \
        # That is, there is no node with two initialzier.
        return self.get_init_node_input(mul) and self.get_init_node_input(add)


class Pattern_4(ONNXTransformer):
    """
    transform
        prev --> BatchNormalization --> next
    to
        prev --> Mul --> Add --> next

    if prev.op_type != Conv
    """

    pattern_to_match = ['BatchNormalization']

    def pattern_matching(self, base_node: onnx.NodeProto) -> List[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        batch_norm = matched_nodes[0]
        if utils.is_op_type(self.find_prev_node(batch_norm.input[0]), ['Conv']):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=_make_new_node_pattern_4(matched_nodes),
            inits_to_add=self.make_new_init(matched_nodes),
            vis_to_add=self.make_new_vi(matched_nodes),
        )

        return batch_norm.input

    def make_new_init(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.TensorProto]:
        (batch_norm,) = matched_nodes
        bn_params = _get_bn_params(batch_norm, self.get_initializer_array)
        multiplier, shifter = _get_multiplier_and_shifter(*bn_params)
        num_features = self.get_value_info_shape(batch_norm.output[0])[0]
        return [
            onnx.numpy_helper.from_array(
                multiplier.reshape(num_features, -1, 1, 1),
                name=batch_norm.output[0] + '_bn_multiplier',
            ),
            onnx.numpy_helper.from_array(
                shifter.reshape(num_features, -1, 1, 1), name=batch_norm.output[0] + '_bn_shifter'
            ),
        ]

    def make_new_vi(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.ValueInfoProto]:
        (batch_norm,) = matched_nodes
        return [
            onnx.helper.make_tensor_value_info(
                batch_norm.output[0] + '_bn_multiplied',
                onnx.TensorProto.FLOAT,
                shape=self.get_value_info_shape(batch_norm.output[0]),
            )
        ]


def _get_bn_params(
    node: onnx.NodeProto, get_init_arr_func: Callable[[str], np.ndarray]
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, float]:
    scale = get_init_arr_func(node.input[1])
    if all(v == 0.0 for v in scale):
        logger.warning('BatchNormalization.scale is a zero tensor: %s', node.input[1])

    B = get_init_arr_func(node.input[2])
    mean = get_init_arr_func(node.input[3])
    var = get_init_arr_func(node.input[4])

    eps = utils.get_attribute(node.attribute, "epsilon", 1e-05)

    return scale, B, mean, var, eps


def _fuse_bn_weight(weight: np.ndarray, multiplier: np.ndarray, axis: int = 0) -> np.ndarray:
    idx = [1, 1, 1]
    idx.insert(axis, -1)
    return weight * multiplier.reshape(idx)


def _fuse_bn_bias(bias: np.ndarray, multiplier: np.ndarray, shifter: np.ndarray):
    return bias * multiplier + shifter


def _get_multiplier_and_shifter(
    scale: np.ndarray, B: np.ndarray, mean: np.ndarray, var: np.ndarray, eps: float
) -> Tuple[np.ndarray, np.ndarray]:
    reciprocal = 1 / np.sqrt(var + eps)
    multiplier = scale * reciprocal
    shifter = -mean * scale * reciprocal + B
    return multiplier, shifter


def _make_bn_fused_node(node: onnx.NodeProto, output_tensor: str) -> List[onnx.NodeProto]:
    assert node.op_type in ['Conv', 'ConvTranspose'], repr(node)
    input_names = [node.input[0], node.input[1] + '_bn_fused']
    if len(node.input) == 3:
        input_names.append(node.input[2] + '_bn_fused')
    else:
        input_names.append(node.output[0] + '_bias_bn_fused')

    return [
        onnx.helper.make_node(
            node.op_type,
            input_names,
            [output_tensor],
            node.name,
            **{attr.name: onnx.helper.get_attribute_value(attr) for attr in node.attribute},
        )
    ]


def _make_bn_fused_init(
    node: onnx.NodeProto,
    multiplier: np.ndarray,
    shifter: np.ndarray,
    get_init_arr_func: Callable[[str], np.ndarray],
) -> List[onnx.TensorProto]:
    assert node.op_type in ['Conv', 'ConvTranspose'], repr(node)
    weight = get_init_arr_func(node.input[1])
    fused_weight = _fuse_bn_weight(weight, multiplier, axis=0 if node.op_type == 'Conv' else 1)
    fused_weight_name = node.input[1] + '_bn_fused'

    if len(node.input) == 3:
        bias = get_init_arr_func(node.input[2])
        fused_bias_name = node.input[2] + '_bn_fused'
    else:
        bias = np.zeros_like(shifter)
        fused_bias_name = node.output[0] + '_bias_bn_fused'
    fused_bias = _fuse_bn_bias(bias, multiplier, shifter)

    return [
        onnx.numpy_helper.from_array(fused_weight, name=fused_weight_name),
        onnx.numpy_helper.from_array(fused_bias, name=fused_bias_name),
    ]


def _make_new_node_pattern_4(matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.NodeProto]:
    (batch_norm,) = matched_nodes
    return [
        onnx.helper.make_node(
            'Mul',
            [batch_norm.input[0], batch_norm.output[0] + '_bn_multiplier'],
            [batch_norm.output[0] + '_bn_multiplied'],
            batch_norm.name,
        ),
        onnx.helper.make_node(
            'Add',
            [batch_norm.output[0] + '_bn_multiplied', batch_norm.output[0] + '_bn_shifter'],
            [batch_norm.output[0]],
            batch_norm.name,
        ),
    ]



from typing import AbstractSet, Dict, Iterable, List, Optional, Sequence, Tuple

import numpy as np
import onnx

from furiosa.quantizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.quantizer.interfaces.transformer import Transformer


# TODO: consider (init, data) input case for MatMul
# related issue: https://github.com/furiosa-ai/furiosa-sdk-private/issues/243
class FuseConv(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:
        for transformer in [
            Pattern_1,
            Pattern_2,
            Pattern_3,
        ]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> MatMul --> Add --> next
    to
        prev --> Unsqueeze --> Conv --> Squeeze --> next

    if 1. MatMul.ndim == 2
       2. MatMul must have exactly one initializer
       3. Add must have exactly one initializer
       4. Add's input with initializer is multidirectional broadcastable to (1, oC)
    """

    pattern_to_match = ['MatMul', 'Add']

    def pattern_matching(self, base_node: onnx.NodeProto) -> List[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=self.make_new_node(matched_nodes),
            inits_to_add=self.make_new_init(matched_nodes),
            vis_to_add=self.make_new_vi(matched_nodes),
        )

        matmul = matched_nodes[0]
        return matmul.input

    def pattern_condition_checker(self, nodes_to_check: Iterable[onnx.NodeProto]) -> bool:
        matmul, add = nodes_to_check
        return (
            self.check_condition_1(matmul)
            and self.check_condition_2(matmul)
            and self.check_condition_2(add)
            and self.check_condition_3(matmul, add)
        )

    def check_condition_1(self, node: onnx.NodeProto) -> bool:
        return len(self.get_value_info_shape(node.output[0])) == 2

    def check_condition_2(self, node: onnx.NodeProto) -> bool:
        return sum(node_input in self.initializer_map for node_input in node.input) == 1

    def check_condition_3(self, node: onnx.NodeProto, node_1: onnx.NodeProto) -> bool:
        bias_tensor = self.get_init_node_input(node_1)
        bias_arr = self.get_initializer_array(bias_tensor)
        oC = self.get_value_info_shape(node.output[0])[1]

        return _is_np_broadcastable(bias_arr, (1, oC))

    def make_new_node(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.NodeProto]:
        matmul, add = matched_nodes
        matmul_data_tensor = self.get_data_node_input(matmul)
        matmul_init_tensor = self.get_init_node_input(matmul)
        add_init_tensor = self.get_init_node_input(add)

        unsqueeze = onnx.helper.make_node(
            'Unsqueeze',
            inputs=[matmul_data_tensor],
            outputs=[matmul.output[0] + '_unsqueezed'],
            name=matmul.output[0] + '_1',
            axes=[2, 3],
        )

        conv = onnx.helper.make_node(
            'Conv',
            inputs=[
                unsqueeze.output[0],
                matmul_init_tensor + '_fused',
                add_init_tensor + '_fused',
            ],
            outputs=[matmul.output[0] + '_fused'],
            name=matmul.output[0] + '_2',
        )

        squeeze = onnx.helper.make_node(
            'Squeeze',
            inputs=[conv.output[0]],
            outputs=[add.output[0]],
            name=matmul.output[0] + '_3',
            axes=[2, 3],
        )
        return [unsqueeze, conv, squeeze]

    def make_new_init(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.TensorProto]:
        matmul, add = matched_nodes

        weight_tensor = self.get_init_node_input(matmul)
        w_arr = self.get_initializer_array(weight_tensor)
        new_w_arr = _matmul_weight_transform(w_arr)
        new_weight = onnx.numpy_helper.from_array(new_w_arr, weight_tensor + '_fused')

        bias_tensor = self.get_init_node_input(add)
        bias_arr = self.get_initializer_array(bias_tensor)
        oC = self.get_value_info_shape(matmul.output[0])[1]
        new_bias_arr = np.broadcast_to(bias_arr, (1, oC)).flatten()
        new_bias = onnx.numpy_helper.from_array(new_bias_arr, bias_tensor + '_fused')

        return [new_weight, new_bias]

    def make_new_vi(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.ValueInfoProto]:
        matmul, _ = matched_nodes
        matmul_data_tensor = self.get_data_node_input(matmul)

        conv_input_vi = onnx.helper.make_tensor_value_info(
            matmul.output[0] + '_unsqueezed',
            onnx.TensorProto.FLOAT,
            self.get_value_info_shape(matmul_data_tensor) + [1, 1],
        )

        conv_output_vi = onnx.helper.make_tensor_value_info(
            matmul.output[0] + '_fused',
            onnx.TensorProto.FLOAT,
            self.get_value_info_shape(matmul.output[0]) + [1, 1],
        )

        return [conv_input_vi, conv_output_vi]


class Pattern_2(ONNXTransformer):
    """
    transform
        prev --> Gemm --> next
    to
        prev --> Unsqueeze --> Conv --> Squeeze --> next
    if 1. Gemm.B must be defined in initializer whereas Gemm.A must not
       2. if Gemm.C is defined, Gemm.C must be an initializer and multidirectional broadcastable to (1, oC)
       3. all of Gemm.input must have onnx.TensorProto.FLOAT dtype
    """

    pattern_to_match = ['Gemm']

    def pattern_matching(self, base_node: onnx.NodeProto) -> List[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=self.make_new_node(matched_nodes),
            inits_to_add=self.make_new_init(matched_nodes),
            vis_to_add=self.make_new_vi(matched_nodes),
        )
        gemm = matched_nodes[0]
        return gemm.input

    def pattern_condition_checker(self, nodes_to_check: Iterable[onnx.NodeProto]) -> bool:
        (gemm,) = nodes_to_check
        return (
            self.check_condition_1(gemm)
            and self.check_condition_2(gemm)
            and self.check_condition_3(gemm)
        )

    def check_condition_1(self, node: onnx.NodeProto) -> bool:
        return node.input[0] not in self.initializer_map and node.input[1] in self.initializer_map

    def check_condition_2(self, node: onnx.NodeProto) -> bool:
        if len(node.input) == 3:
            if node.input[2] in self.initializer_map:
                array = self.get_initializer_array(node.input[2])
                oC = self.get_value_info_shape(node.output[0])[1]

                return _is_np_broadcastable(array, (1, oC))
            # returns False since Gemm.C has no initializer to be fused
            return False
        # always returns True if Gemm.C is not defined.
        return True

    def check_condition_3(self, node: onnx.NodeProto) -> bool:
        return all(
            self.get_value_info_dtype(tensor) == onnx.TensorProto.FLOAT for tensor in node.input
        )

    def make_new_node(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.NodeProto]:
        (gemm,) = matched_nodes
        input_tensor, weight_tensor, bias_tensor = _get_gemm_inputs(gemm, set(self.initializer_map))
        attrs = _get_gemm_attrs(gemm)

        new_nodes = []
        unsqueeze_input = input_tensor
        input_idx = _get_input_index(input_tensor, gemm)
        if _needs_gemm_transpose(input_idx, attrs):
            unsqueeze_input += '_transposed'
            transpose = onnx.helper.make_node(
                'Transpose',
                inputs=[input_tensor],
                outputs=[unsqueeze_input],
                name=gemm.output[0] + '_0',
            )
            new_nodes.append(transpose)

        unsqueeze = onnx.helper.make_node(
            'Unsqueeze',
            inputs=[unsqueeze_input],
            outputs=[gemm.output[0] + '_unsqueezed'],
            name=gemm.output[0] + '_1',
            axes=[2, 3],
        )

        conv_inputs = [unsqueeze.output[0], weight_tensor + '_fused']
        if bias_tensor is not None:
            conv_inputs.append(bias_tensor + '_fused')
        conv = onnx.helper.make_node(
            'Conv',
            inputs=conv_inputs,
            outputs=[gemm.output[0] + '_fused'],
            name=gemm.output[0] + '_2',
        )

        squeeze = onnx.helper.make_node(
            'Squeeze',
            inputs=[conv.output[0]],
            outputs=[gemm.output[0]],
            name=gemm.output[0] + '_3',
            axes=[2, 3],
        )

        new_nodes.extend([unsqueeze, conv, squeeze])
        return new_nodes

    def make_new_init(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.TensorProto]:
        (gemm,) = matched_nodes
        _, weight_tensor, bias_tensor = _get_gemm_inputs(gemm, set(self.initializer_map))
        attrs = _get_gemm_attrs(gemm)

        new_inits = []
        w_arr = self.get_initializer_array(weight_tensor) * attrs['alpha']
        new_w_arr = _gemm_weight_transfrom(w_arr, attrs)
        new_w_init = onnx.numpy_helper.from_array(new_w_arr, weight_tensor + '_fused')
        new_inits.append(new_w_init)

        if bias_tensor:
            b_arr = self.get_initializer_array(bias_tensor)
            oC = self.get_value_info_shape(gemm.output[0])[1]
            new_b_arr = np.broadcast_to(b_arr, (1, oC)).flatten() * attrs['beta']
            new_b_init = onnx.numpy_helper.from_array(new_b_arr, bias_tensor + '_fused')
            new_inits.append(new_b_init)

        return new_inits

    def make_new_vi(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.ValueInfoProto]:
        (gemm,) = matched_nodes
        input_tensor, _, _ = _get_gemm_inputs(gemm, set(self.initializer_map))

        new_vis = []
        attrs = _get_gemm_attrs(gemm)
        input_idx = _get_input_index(input_tensor, gemm)
        if _needs_gemm_transpose(input_idx, attrs):
            transpose_output_vi = onnx.helper.make_tensor_value_info(
                input_tensor + '_transposed',
                onnx.TensorProto.FLOAT,
                self.get_value_info_shape(input_tensor)[::-1],
            )
            new_vis.append(transpose_output_vi)
            # input shape also should be transposed
            input_shape = self.get_value_info_shape(input_tensor)[::-1] + [1, 1]
        else:
            input_shape = self.get_value_info_shape(input_tensor) + [1, 1]

        conv_input_vi = onnx.helper.make_tensor_value_info(
            gemm.output[0] + '_unsqueezed',
            onnx.TensorProto.FLOAT,
            input_shape,
        )

        conv_output_vi = onnx.helper.make_tensor_value_info(
            gemm.output[0] + '_fused',
            onnx.TensorProto.FLOAT,
            self.get_value_info_shape(gemm.output[0]) + [1, 1],
        )
        new_vis.extend([conv_input_vi, conv_output_vi])
        return new_vis


class Pattern_3(ONNXTransformer):
    """
    transform
        prev --> Conv --> Add --> next
    to
        prev --> Conv --> next
    if  1. len(Conv.input) == 2 or (len(Conv.input) == 3 and Conv.input[2] has initializer)
        2. Add has only one initializer
        3. Add's input with initializer is multidirectional broadcastable to (1, oC, 1, 1)
    """

    pattern_to_match = ['Conv', 'Add']

    def pattern_matching(self, base_node: onnx.NodeProto) -> List[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=self.make_new_node(matched_nodes),
            inits_to_add=self.make_new_init(matched_nodes),
        )
        conv = matched_nodes[0]
        return conv.input

    def pattern_condition_checker(self, nodes_to_check: Iterable[onnx.NodeProto]) -> bool:
        conv, add = nodes_to_check
        return (
            self.check_condition_1(conv)
            and self.check_condition_2(add)
            and self.check_condition_3(conv, add)
        )

    def check_condition_1(self, node: onnx.NodeProto) -> bool:
        return len(node.input) == 2 or node.input[2] in self.initializer_map

    def check_condition_2(self, node: onnx.NodeProto) -> bool:
        return sum(node_input in self.initializer_map for node_input in node.input) == 1

    def check_condition_3(self, node: onnx.NodeProto, node_1: onnx.NodeProto) -> bool:
        bias_tensor = self.get_init_node_input(node_1)
        bias_arr = self.get_initializer_array(bias_tensor)
        oC = self.get_value_info_shape(node.output[0])[1]

        return _is_np_broadcastable(bias_arr, (1, oC, 1, 1))

    def make_new_node(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.NodeProto]:
        conv, add = matched_nodes
        return [
            onnx.helper.make_node(
                'Conv',
                inputs=[*conv.input[:2], self.get_init_node_input(add) + '_fused'],
                outputs=[add.output[0]],
                name=conv.name,
                **{attr.name: onnx.helper.get_attribute_value(attr) for attr in conv.attribute},
            )
        ]

    def make_new_init(self, matched_nodes: Iterable[onnx.NodeProto]) -> List[onnx.TensorProto]:
        conv, add = matched_nodes
        bias_tensor_name = self.get_init_node_input(add)
        bias_array = self.get_initializer_array(bias_tensor_name)
        oC = self.get_value_info_shape(conv.output[0])[1]
        bias_array = np.broadcast_to(bias_array, (1, oC, 1, 1)).flatten()
        if len(conv.input) == 3:
            bias_array += self.get_initializer_array(conv.input[2]).flatten()
        return [onnx.numpy_helper.from_array(bias_array, bias_tensor_name + '_fused')]


# TODO: fuse Conv + Mul into Conv
# class Pattern_4(Pattern_1):
#    """
#    transform
#        prev --> Conv --> Mul --> next
#    to
#        prev --> Conv --> next
#    if ...
#    """


def _is_np_broadcastable(array: np.ndarray, to_shape: Sequence[int]) -> bool:
    try:
        np.broadcast_to(array, to_shape)
        return True
    except ValueError:
        return False


def _matmul_weight_transform(weight: np.ndarray) -> np.ndarray:
    c, oC = weight.shape
    new_w_arr = weight.transpose().reshape(oC, c, 1, 1)
    return new_w_arr


def _gemm_weight_transfrom(weight: np.ndarray, attrs: Dict) -> np.ndarray:
    if attrs['transB'] != 0:
        weight = weight.transpose()
    return _matmul_weight_transform(weight)


def _get_gemm_inputs(
    node: onnx.NodeProto, init_keys: AbstractSet[str]
) -> Tuple[str, str, Optional[str]]:
    assert node.op_type == "Gemm", repr(node)

    for node_input in node.input[:2]:
        if node_input in init_keys:
            weight_tensor = node_input
        else:
            input_tensor = node_input

    bias_tensor = node.input[2] if len(node.input) == 3 else None

    return input_tensor, weight_tensor, bias_tensor


def _get_input_index(input_tensor: str, node: onnx.NodeProto) -> int:
    return list(node.input).index(input_tensor)


def _needs_gemm_transpose(input_idx: int, attrs: Dict) -> bool:
    assert input_idx in (0, 1)
    return attrs['transA' if input_idx == 0 else 'transB'] != 0


def _get_gemm_attrs(node: onnx.NodeProto) -> Dict:
    assert node.op_type == "Gemm", repr(node)

    attrs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in node.attribute}

    return {
        'alpha': attrs.get('alpha', 1.0),
        'beta': attrs.get('beta', 1.0),
        'transA': attrs.get('transA', 0),
        'transB': attrs.get('transB', 0),
    }



import onnx
from onnx import checker
import onnxruntime as ort


def check_model(model: onnx.ModelProto, check_runnable: bool = True) -> None:
    """
    Check if model's well-defined and executable on onnxruntime
    """
    # TODO After collecting possible errors,
    #  pass through only if all error messages are "No opset import for domain 'com.microsoft'".
    #  The code below is only to see the first error encountered.
    acceptable_error_msg = [
        "No opset import for domain 'com.microsoft'",
        'No Op registered for LayerNormalization with domain_version of 12',
    ]
    try:
        checker.check_model(model)
    except checker.ValidationError as e:
        if str(e).split("==>", maxsplit=1)[0].rstrip() in acceptable_error_msg:
            pass
        else:
            checker.check_model(model)

    if check_runnable:
        ort.InferenceSession(model.SerializeToString())



import itertools
import logging
from typing import Any, Callable, Iterable, List, Optional, TypeVar

import onnx
from onnx.helper import make_model, make_opsetid, make_tensor_value_info

from furiosa.quantizer.frontend.onnx import __DOMAIN__, __OPSET_VERSION__
from furiosa.quantizer.frontend.onnx.quantizer.utils import __PRODUCER__

logger = logging.getLogger('Furiosa-Quantizer')
logging.basicConfig(level=logging.INFO)


def name_nodes(model):
    for idx, node in enumerate(model.graph.node):
        node.name = f'{node.op_type}_{idx}'

    return model


def eliminate_unused_initializer(model: onnx.ModelProto) -> onnx.ModelProto:
    """
    This function eliminates every initializers not used by node input,
    regardless of any graph fields they are defined in.
    """
    node_input_names = set(tensor_name for node in model.graph.node for tensor_name in node.input)
    used_initializers = [
        tensor for tensor in model.graph.initializer if tensor.name in node_input_names
    ]
    del model.graph.initializer[:]
    model.graph.initializer.extend(used_initializers)
    return model


def eliminate_unused_input(model):
    node_input_names = set(tensor_name for node in model.graph.node for tensor_name in node.input)
    graph_output_names = set(value_info.name for value_info in model.graph.output)
    used_inputs = [
        value_info
        for value_info in model.graph.input
        if value_info.name in node_input_names or value_info.name in graph_output_names
    ]
    del model.graph.input[:]
    model.graph.input.extend(used_inputs)
    return model


def eliminate_unused_output(model):
    node_output_names = set(tensor_name for node in model.graph.node for tensor_name in node.output)
    graph_input_names = set(value_info.name for value_info in model.graph.input)
    used_outputs = [
        value_info
        for value_info in model.graph.output
        if value_info.name in node_output_names or value_info.name in graph_input_names
    ]
    del model.graph.output[:]
    model.graph.output.extend(used_outputs)
    return model


def eliminate_unused_value_info(model):
    node_output_names = set(tensor_name for node in model.graph.node for tensor_name in node.output)
    graph_output_names = set(value_info.name for value_info in model.graph.output)
    used_value_infos = [
        value_info
        for value_info in model.graph.value_info
        if value_info.name in node_output_names and value_info.name not in graph_output_names
    ]
    del model.graph.value_info[:]
    model.graph.value_info.extend(used_value_infos)
    return model


def eliminate_unused_protos(model):
    funcs = [
        eliminate_unused_initializer,
        eliminate_unused_input,
        eliminate_unused_output,
        eliminate_unused_value_info,
    ]

    for func in funcs:
        model = func(model)

    return model


def eliminate_initializer_from_graph_input(model: onnx.ModelProto) -> onnx.ModelProto:
    initializer = set(init.name for init in model.graph.initializer)
    graph_input = [
        value_info for value_info in model.graph.input if value_info.name not in initializer
    ]
    del model.graph.input[:]
    model.graph.input.extend(graph_input)

    return model


def include_initializer_to_graph_input(model: onnx.ModelProto) -> onnx.ModelProto:
    input_value_names = [inp.name for inp in model.graph.input]
    for init in model.graph.initializer:
        if init.name not in input_value_names:
            value_info = make_tensor_value_info(init.name, init.data_type, init.dims)
            model.graph.input.append(value_info)
            # do not append duplicated initializer to graph input
            input_value_names.append(init.name)
    return model


def rebuild_model(
    model: onnx.ModelProto,
    new_nodes: List[onnx.NodeProto],
    eliminate: bool = True,
    renaming: bool = True,
):
    # remove all nodes and re-make model.graph based on newly given nodes.
    del model.graph.node[:]
    model.graph.node.extend(new_nodes)
    default_opset = make_opsetid(__DOMAIN__, __OPSET_VERSION__)
    model = make_model(model.graph, opset_imports=[default_opset])

    # eliminate all unused protos such as initializer, input, output, and value_info.
    if eliminate:
        model = eliminate_initializer_from_graph_input(model)
        model = eliminate_unused_protos(model)

    # rename node.name
    if renaming:
        model = name_nodes(model)
    model.producer_name = __PRODUCER__
    return model


def fix_batch_size_as_one(model):
    """
    fix batch_size = 1 if dim_param is given.
    """
    for value_info in model.graph.input:
        try:
            batch_dim = value_info.type.tensor_type.shape.dim[0]
        except IndexError:
            continue

        if batch_dim.dim_param:
            logger.info(
                "Dynamic batch size is detected at input_name: %s. "
                "Fix batch_size=1 for valid shape inference.",
                value_info.name,
            )
            value_info.type.tensor_type.shape.dim[0].dim_value = 1

    return model


def make_initializer_name_unique(model):
    # Renames Operators' initializers, if necessary, to make their names unique
    initializer = {init.name: init for init in model.graph.initializer}
    del model.graph.initializer[:]
    for node in model.graph.node:
        for idx, node_input in enumerate(node.input):
            if node_input not in initializer:
                continue

            tensor = onnx.TensorProto()
            tensor.CopyFrom(initializer[node_input])
            # HACK: This attempts to give the initializer a new unique name.
            # Although it is unlikely, there is a possibility that the new
            # name is already occupied by a tensor in the model.
            tensor.name = f"{node_input}_{node.output[0]}_{idx}"
            node.input[idx] = tensor.name
            model.graph.initializer.append(tensor)

    return model


T = TypeVar('T')


def fixed_point(x: T, functions: Iterable[Callable[[T], T]]) -> T:
    while True:
        init = x
        for func in functions:
            x = func(x)
        if x == init:
            return x


def get_attribute(
    attrs: Iterable[onnx.AttributeProto], attr_name: str, default: Optional[Any] = None
) -> Any:
    return next(
        (onnx.helper.get_attribute_value(attr) for attr in attrs if attr.name == attr_name), default
    )


def make_unhashables_unique(values):
    seen = []
    for v in values:
        if v not in seen:
            seen.append(v)

    return seen


def is_op_type(op_type: str, target_op_types: Iterable[str]) -> bool:
    return op_type in target_op_types


def check_value_info(model: onnx.ModelProto) -> None:
    initializer = {init.name: init for init in model.graph.initializer}
    value_info = {
        vi.name: vi
        for vi in itertools.chain(model.graph.value_info, model.graph.input, model.graph.output)
    }
    tensor_names = set(
        tensor_name for node in model.graph.node for tensor_name in (*node.input, *node.output)
    )
    for name in tensor_names:
        if (
            name in initializer or not name
        ):  # empty name indicates that optional input is unspecified
            continue

        if name not in value_info:
            raise ValueError(
                f'value_info of {name} is missing. Optimize model before quantization.'
            )
        try:
            if not value_info[name].type.tensor_type.HasField('elem_type'):
                raise ValueError(
                    f'elem_type of {name} in value_info is missing. Optimize model before quantization, or shape inference failed.'
                )
            if value_info[name].type.tensor_type.elem_type != onnx.TensorProto.FLOAT:
                logger.warning(
                    'elem_type of %s(%s) is not FLOAT: Model might be already quantized.',
                    name,
                    onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[
                        value_info[name].type.tensor_type.elem_type
                    ],
                )
            if not value_info[name].type.tensor_type.HasField(
                'shape'
            ):  # when shape inference failed, shape field does not exist, unlike empty shape.dim array for scalar.
                raise ValueError(
                    f'shape of {name} in value_info is missing. Optimize model before quantization, or shape inference failed.'
                )
        except AttributeError as e:
            raise AttributeError(
                f'{e} (ValueInfoProto is incomplete. Optimize model before quantization, or shape inference failed.)'
            ) from None



class HeightWidth:
    def __init__(self, height: int, width: int):
        self.height = height
        self.width = width


class Padding:
    def __init__(self, top: int, bottom: int, left: int, right: int):
        self.top = top
        self.bottom = bottom
        self.left = left
        self.right = right


class HorizontalPadding:
    def __init__(self, first: int, second: int):
        self.first = first
        self.second = second
